import dataclasses
import pathlib
import typing
from enum import Enum
from typing import Callable, List, Optional, Union

import lark
import typer

from rbx import console
from rbx.box import checkers, package
from rbx.box.schema import Checker, CodeItem, ExpectedOutcome
from rbx.box.solutions import expand_solutions
from rbx.grading.steps import CheckerResult, Outcome, RunLog, TestcaseLog

LARK_GRAMMAR = r"""
// A bunch of words
start: disjunction

disjunction: conjunction | disjunction _OR conjunction

conjunction: _atom | conjunction _AND _atom

_atom: logical | "(" disjunction ")" | negation
negation: _NOT "(" disjunction ")"

// Expressions
logical: eval matcher expected_outcome -> matching
       | eval equality (eval | outcome) -> equating
       | eval -> eval_only

eval: "[" solution checking? "]" | solution

// Eval
solution: _solution_filename | WILDCARD
checking: "ON"i (checking_mode? checker | ":nil")
checking_mode: MODE ":"
MODE: "2" | "3"
checker: _filename | WILDCARD

// Outcomes
expected_outcome: CNAME
outcome: CNAME

// Operators
matcher: MATCHES | NOT_MATCHES
equality: EQUALS | NOT_EQUALS

MATCHES: "~"
NOT_MATCHES: "!~"
EQUALS: "=="
NOT_EQUALS: "!="
_OR: "||"
_AND: "&&"
_NOT: "!"
WILDCARD: "$"

// File name
_filename: FILENAME | "\"" FILENAME "\""
_solution_filename: _filename | "@" _filename
FILENAME: /[\/A-Za-z0-9\-_\.]/+

// Names (Variables)
LCASE_LETTER: "a".."z"
UCASE_LETTER: "A".."Z"
DIGIT: "0".."9"
LETTER: UCASE_LETTER | LCASE_LETTER
WORD: LETTER+
CNAME: ("_"|LETTER) ("_"|LETTER|DIGIT|"+")*

%ignore " "
"""

LARK_PARSER = lark.Lark(LARK_GRAMMAR)


class CheckingMode(Enum):
    THREE_WAY = 0
    TWO_WAY = 1


@dataclasses.dataclass(frozen=True)
class FinderChecker:
    path: str
    mode: CheckingMode


@dataclasses.dataclass(frozen=True)
class FinderCall:
    solution: str
    checker: Optional[FinderChecker]


@dataclasses.dataclass(frozen=True)
class FinderSolutionResult:
    output_path: pathlib.Path
    stderr_path: Optional[pathlib.Path] = None
    run_log: Optional[RunLog] = None


@dataclasses.dataclass(frozen=True)
class FinderResult:
    solution: str
    outcome: Outcome
    checker: Optional[FinderChecker]

    # Auxiliary information.
    solution_log: Optional[TestcaseLog] = None
    checker_result: Optional[CheckerResult] = None


@dataclasses.dataclass(frozen=True)
class FinderOutcome:
    truth_value: bool
    results: List[FinderResult]


def or_outcome(a: FinderOutcome, b: FinderOutcome) -> FinderOutcome:
    return FinderOutcome(
        truth_value=a.truth_value or b.truth_value, results=a.results + b.results
    )


def and_outcome(a: FinderOutcome, b: FinderOutcome) -> FinderOutcome:
    return FinderOutcome(
        truth_value=a.truth_value and b.truth_value, results=a.results + b.results
    )


def get_checking_mode_from_string(mode: Optional[str]) -> CheckingMode:
    if not mode:
        return CheckingMode.THREE_WAY
    if mode == '2':
        return CheckingMode.TWO_WAY
    return CheckingMode.THREE_WAY


def _get_main_checker() -> Optional[str]:
    checker = package.get_checker()
    if checker is None:
        return None
    return str(checker.path)


def _get_main_solution() -> Optional[str]:
    sol = package.get_main_solution()
    if sol is None:
        return None
    return str(sol.path)


def _get_default_checker_for_finder() -> Optional[FinderChecker]:
    main_checker = _get_main_checker()
    if main_checker is None:
        return None
    return FinderChecker(path=main_checker, mode=CheckingMode.THREE_WAY)


def _get_solution_from_token(token: lark.Token) -> str:
    path = str(token)
    if path == '$':
        main_path = _get_main_solution()
        assert main_path is not None
        return main_path
    return path


def _get_checker_from_token(token: lark.Token) -> str:
    path = str(token)
    if path == '$':
        main_path = _get_main_checker()
        assert main_path is not None
        return main_path
    return path


def _get_eval_checker(eval: lark.ParseTree) -> Optional[FinderChecker]:
    checking_nodes = list(eval.find_data('checking'))
    if not checking_nodes:
        return _get_default_checker_for_finder()
    (checking,) = checking_nodes

    if not checking.children:
        # Checking is nil
        return None

    if len(checking.children) == 1:
        checker = typing.cast(lark.ParseTree, checking.children[0])
        return FinderChecker(
            path=_get_checker_from_token(typing.cast(lark.Token, checker.children[0])),
            mode=CheckingMode.THREE_WAY,
        )

    mode = typing.cast(lark.ParseTree, checking.children[0])
    checker = typing.cast(lark.ParseTree, checking.children[1])

    return FinderChecker(
        path=_get_checker_from_token(typing.cast(lark.Token, checker.children[0])),
        mode=get_checking_mode_from_string(
            typing.cast(lark.Token, mode.children[0]).value
        ),
    )


def get_all_solutions(tree: lark.ParseTree) -> List[str]:
    solution_nodes = tree.find_data('solution')
    res = set(
        [
            _get_solution_from_token(typing.cast(lark.Token, node.children[0]))
            for node in solution_nodes
        ]
    )

    if needs_expected_output(tree):
        main_solution = package.get_main_solution()
        assert main_solution is not None
        res.add(str(main_solution.path))
    return list(res)


def get_all_solution_items(tree: lark.ParseTree) -> List[CodeItem]:
    solution_names = get_all_solutions(tree)
    res = typing.cast(List[CodeItem], expand_solutions(solution_names))

    main_solution = package.get_main_solution()
    if main_solution is None:
        return res

    for i, sol in enumerate(res):
        if main_solution.path == sol.path:
            res[i], res[0] = res[0], res[i]
    return res


def _get_all_finder_checkers(tree: lark.ParseTree) -> List[FinderChecker]:
    eval_nodes = tree.find_data('eval')
    res = []

    for eval_node in eval_nodes:
        finder_checker = _get_eval_checker(eval_node)
        if finder_checker is not None:
            res.append(finder_checker)

    return res


def get_all_checkers(tree: lark.ParseTree) -> List[str]:
    return [finder_checker.path for finder_checker in _get_all_finder_checkers(tree)]


def get_all_checker_items(tree: lark.ParseTree) -> List[Checker]:
    checker_names = get_all_checkers(tree)
    res = []

    for checker_name in checker_names:
        main_checker = package.get_checker()
        if str(main_checker.path) == checker_name:
            res.append(main_checker)
            continue
        res.append(
            Checker(
                path=pathlib.Path(checker_name),
                language=None,
                compilationFiles=None,
            )
        )
    return res


def needs_expected_output(tree: lark.ParseTree) -> bool:
    finder_checkers = _get_all_finder_checkers(tree)
    for finder_checker in finder_checkers:
        if finder_checker.mode == CheckingMode.THREE_WAY:
            return True
    return False


def validate(tree: lark.ParseTree):
    if needs_expected_output(tree):
        if package.get_main_solution() is None:
            console.console.print(
                '[error]Finder expression requires three-way checking, but problem has no main solution.[/error]'
            )
            console.console.print(
                'Either provide an ACCEPTED solution at your problem.rbx.yml, or use two-way checking in your finder expression by providing the `2:` parameter before the checker name.'
            )
            raise typer.Exit(1)

    all_checkers = get_all_checkers(tree)
    for checker in all_checkers:
        if not checkers.is_valid_checker(pathlib.Path(checker)):
            console.console.print(
                f'[error]Finder expression references non-existing checker [item]{checker}[/item][/error]'
            )
            raise typer.Exit(1)

    all_solutions = get_all_solutions(tree)
    for solution in all_solutions:
        if not pathlib.Path(solution).is_file():
            console.console.print(
                f'[error]Finder expression references non-existing solution [item]{solution}[/item][/error]'
            )
            raise typer.Exit(1)


@lark.v_args(inline=True)
class FinderTreeRunner(lark.Transformer):
    def __init__(
        self,
        runner: Callable[[FinderCall], FinderResult],
        eval_only_outcome: ExpectedOutcome = ExpectedOutcome.INCORRECT,
    ):
        self.run_fn = runner
        self.eval_only_outcome = eval_only_outcome

    def solution(self, token: lark.Token) -> str:
        return _get_solution_from_token(token)

    def outcome(self, token: lark.Token) -> Outcome:
        try:
            outcome = Outcome(token.value)
        except ValueError:
            try:
                expected_outcome = self.expected_outcome(token)
            except ValueError:
                raise ValueError(f'"{token.value}" is not a valid Outcome.') from None
            outcomes = expected_outcome.get_matches()
            if len(outcomes) != 1:
                raise ValueError(
                    f'"{token.value}" is not a valid Outcome. You are trying to specify an ExpectedOutcome, instead of a single Outcome.'
                ) from None
            return outcomes[0]
        return outcome

    def expected_outcome(self, token: lark.Token) -> ExpectedOutcome:
        return ExpectedOutcome(token.value)

    def matcher(self, op: lark.Token) -> bool:
        return op.value == '~'

    def equality(self, op: lark.Token) -> bool:
        return op.value == '=='

    @lark.v_args(inline=False, tree=True)
    def eval(self, tree: lark.ParseTree) -> FinderResult:
        solution = typing.cast(str, tree.children[0])
        checker: Optional[FinderChecker] = _get_eval_checker(tree)

        call = FinderCall(solution, checker=checker)
        return self.run_fn(call)

    def matching(
        self,
        eval_result: FinderResult,
        is_positive: bool,
        expected_outcome: ExpectedOutcome,
    ) -> FinderOutcome:
        truth_value = expected_outcome.match(eval_result.outcome)
        if not is_positive:
            truth_value = not truth_value

        return FinderOutcome(truth_value=truth_value, results=[eval_result])

    def equating(
        self,
        eval_result: FinderResult,
        is_positive: bool,
        result_or_outcome: Union[FinderResult, Outcome],
    ) -> FinderOutcome:
        results = [eval_result]
        truth_value = True

        if isinstance(result_or_outcome, Outcome):
            outcome: Outcome = result_or_outcome
            truth_value = eval_result.outcome == outcome
        else:
            result: FinderResult = result_or_outcome
            truth_value = eval_result.outcome == result.outcome
            results.append(result)

        if not is_positive:
            truth_value = not truth_value

        return FinderOutcome(truth_value=truth_value, results=results)

    def eval_only(self, eval_result: FinderResult) -> FinderOutcome:
        return self.matching(
            eval_result, is_positive=True, expected_outcome=self.eval_only_outcome
        )

    def negation(self, value: FinderOutcome) -> FinderOutcome:
        return dataclasses.replace(value, truth_value=not value.truth_value)

    @lark.v_args(inline=False)
    def conjunction(self, values: List[FinderOutcome]) -> FinderOutcome:
        res = FinderOutcome(truth_value=True, results=[])
        for value in values:
            res = and_outcome(res, value)
        return res

    @lark.v_args(inline=False)
    def disjunction(self, values: List[FinderOutcome]) -> FinderOutcome:
        res = FinderOutcome(truth_value=False, results=[])
        for value in values:
            res = or_outcome(res, value)
        return res

    def start(self, value: FinderOutcome) -> FinderOutcome:
        return value


def parse(expression: str) -> lark.ParseTree:
    tree = LARK_PARSER.parse(expression)
    validate(tree)
    return tree



import os
import pathlib
from typing import Any, Optional

from rbx import utils
from rbx.box import setter_config


def ref(text: Any) -> str:
    return f'[item]{text}[/item]'


def href(
    url: os.PathLike[str],
    text: Optional[str] = None,
    style: str = 'item',
    hyperlink: bool = True,
) -> str:
    custom_text = False
    if text is None:
        text = str(url)
    else:
        custom_text = True

    if not hyperlink:
        return f'[{style}]{text}[/{style}]'

    if not custom_text:
        if not setter_config.get_setter_config().hyperlinks:
            return f'[{style}]{text}[/{style}]'
        if os.environ.get('TERM') in ['vscode']:
            return f'[{style}]{text}[/{style}]'

    if isinstance(url, pathlib.Path):
        url = utils.abspath(url)

    url_str = str(url)
    if pathlib.Path(url_str).exists():
        url_str = f'file://{url_str}'
    return f'[{style}][link={url_str}]{text}[/link][/{style}]'


def get_formatted_memory(memory_in_bytes: int, mib_decimal_places: int = 0) -> str:
    if memory_in_bytes < 1024 * 1024:
        if memory_in_bytes < 1024:
            return f'{memory_in_bytes} B'
        return f'{memory_in_bytes / 1024:.0f} KiB'
    return f'{memory_in_bytes / (1024 * 1024):.{mib_decimal_places}f} MiB'


def get_formatted_time(time_in_ms: int) -> str:
    return f'{time_in_ms} ms'


def get_formatted_time_in_seconds(time_in_seconds: float) -> str:
    return f'{time_in_seconds:.1f} s'



import pathlib
import shlex
from typing import Dict, Iterable, List, Optional, Set, Tuple

import typer
from pydantic import BaseModel

from rbx import console
from rbx.box import checkers, package
from rbx.box.code import SanitizationLevel, compile_item, run_item
from rbx.box.fields import Primitive
from rbx.box.schema import CodeItem
from rbx.box.testcase_extractors import (
    GenerationMetadata,
    GenerationTestcaseEntry,
    extract_generation_testcases_from_groups,
    get_generation_metadata_markup,
)
from rbx.box.testcase_utils import TestcaseEntry
from rbx.grading.judge.sandbox import SandboxBase
from rbx.grading.steps import (
    DigestHolder,
    DigestOrDest,
    DigestOrSource,
    GradingFileOutput,
    Outcome,
)
from rbx.utils import StatusProgress

HitBounds = Dict[str, Tuple[bool, bool]]


class TestcaseValidationInfo(BaseModel):
    validator: CodeItem
    testcase: Optional[TestcaseEntry]
    generation_metadata: Optional[GenerationMetadata]
    path: pathlib.Path
    ok: bool
    hit_bounds: HitBounds
    message: Optional[str] = None

    def href(self) -> str:
        return CodeItem(path=self.path).href()


def _compile_validator(validator: CodeItem) -> str:
    try:
        digest = compile_item(validator, sanitized=SanitizationLevel.PREFER)
    except Exception:
        console.console.print(
            f'[error]Failed compiling validator {validator.href()}.[/error]'
        )
        # console.console.print_exception()
        raise
    return digest


def _bounds_or(lhs: Tuple[bool, bool], rhs: Tuple[bool, bool]) -> Tuple[bool, bool]:
    return (lhs[0] or rhs[0], lhs[1] or rhs[1])


def _process_bounds(log: str) -> HitBounds:
    bounds: HitBounds = {}
    for line in log.splitlines():
        items = line.split(':')
        if len(items) != 2:
            continue
        k, v = items
        v = v.strip()

        if 'constant-bounds' in k:
            continue

        hit = ('min-value-hit' in v, 'max-value-hit' in v)
        if k not in bounds:
            bounds[k] = hit
            continue
        bounds[k] = _bounds_or(bounds[k], hit)
    return bounds


def _merge_hit_bounds(hit_bounds: Iterable[HitBounds]) -> HitBounds:
    res: HitBounds = {}
    for hb in hit_bounds:
        for k, hit in hb.items():
            if k not in res:
                res[k] = hit
                continue
            res[k] = _bounds_or(res[k], hit)
    return res


def _has_group_specific_validator() -> bool:
    pkg = package.find_problem_package_or_die()

    return any(group.validator is not None for group in pkg.testcases)


async def _validate_testcase(
    testcase: pathlib.Path,
    validator: CodeItem,
    validator_digest: str,
    vars: Optional[Dict[str, Primitive]] = None,
    group: Optional[str] = None,
) -> Tuple[bool, Optional[str], HitBounds]:
    vars = vars or {}
    # TODO: check if needs to do some escaping
    var_args = [f'--{k}={v}' for k, v in vars.items()]
    var_args.extend(['--testOverviewLogFileName', 'validator.log'])
    if group is not None:
        var_args.extend(['--group', group])

    message_digest = DigestHolder()
    log_digest = DigestHolder()
    run_log = await run_item(
        validator,
        DigestOrSource.create(validator_digest),
        stdin=DigestOrSource.create(testcase),
        stderr=DigestOrDest.create(message_digest),
        outputs=[
            GradingFileOutput(
                src=pathlib.Path('validator.log'),
                digest=log_digest,
                optional=True,
            )
        ],
        extra_args=shlex.join(var_args) if var_args else None,
    )

    message = package.get_digest_as_string(message_digest.value or '')
    if (
        run_log is not None
        and run_log.exitcode != 0
        and run_log.exitstatus != SandboxBase.EXIT_NONZERO_RETURN
    ):
        console.console.print(
            f'[error]Validator {validator.href()} failed unexpectedly.[/error]'
        )
        console.console.print(f'[error]Summary:[/error] {run_log.get_summary()}')
        console.console.print(f'[error]Message:[/error] {message}')
        console.console.print(f'[error]Testcase:[/error] {testcase}')
        raise typer.Exit(1)

    log_overview = ''
    if log_digest.value is not None:
        log_overview = package.get_digest_as_string(log_digest.value or '')
    return (
        run_log is not None and run_log.exitcode == 0,
        message,
        _process_bounds(log_overview or ''),
    )


async def validate_file(
    testcase: pathlib.Path,
    validator: CodeItem,
    validator_digest: str,
    group: Optional[str] = None,
) -> Tuple[bool, Optional[str], HitBounds]:
    pkg = package.find_problem_package_or_die()
    return await _validate_testcase(
        testcase, validator, validator_digest, vars=pkg.expanded_vars, group=group
    )


def compile_main_validator() -> Optional[Tuple[CodeItem, str]]:
    pkg = package.find_problem_package_or_die()
    if pkg.validator is None:
        return None

    return pkg.validator, _compile_validator(pkg.validator)


async def validate_one_off(
    testcase: pathlib.Path,
    validators: List[CodeItem],
    validator_digests: Dict[str, str],
    generation_metadata: Optional[GenerationMetadata] = None,
    testcase_entry: Optional[TestcaseEntry] = None,
) -> List[TestcaseValidationInfo]:
    res = []
    for validator in validators:
        validator_digest = validator_digests.get(str(validator.path))
        if validator_digest is None:
            console.console.print(
                f'[warning]Validator {validator.href()} not compiled, skipping validation.[/warning]'
            )
            continue
        ok, message, _ = await validate_file(
            testcase,
            validator,
            validator_digest,
            group=testcase_entry.group if testcase_entry is not None else None,
        )
        res.append(
            TestcaseValidationInfo(
                validator=validator,
                testcase=testcase_entry,
                generation_metadata=generation_metadata,
                path=testcase,
                ok=ok,
                hit_bounds={},
                message=message,
            )
        )
    return res


def compile_validators(
    validators: List[CodeItem],
    progress: Optional[StatusProgress] = None,
) -> Dict[str, str]:
    validator_to_compiled_digest = {}

    validator_to_compiled_digest = {}

    for validator in validators:
        if str(validator.path) in validator_to_compiled_digest:
            continue

        if progress:
            progress.update(f'Compiling validator {validator.href()}...')
        validator_to_compiled_digest[str(validator.path)] = _compile_validator(
            validator
        )

    return validator_to_compiled_digest


def compile_validators_for_entries(
    validation_entries: List[GenerationTestcaseEntry],
    progress: Optional[StatusProgress] = None,
) -> Dict[str, str]:
    validators = []

    for entry in validation_entries:
        if entry.validator is not None:
            validators.append(entry.validator)
        validators.extend(entry.extra_validators)

    return compile_validators(validators, progress=progress)


def compile_output_validators_for_entries(
    validation_entries: List[GenerationTestcaseEntry],
    progress: Optional[StatusProgress] = None,
) -> Dict[str, str]:
    validators = []

    for entry in validation_entries:
        validators.extend(entry.output_validators)

    return compile_validators(validators, progress=progress)


async def validate_testcases(
    progress: Optional[StatusProgress] = None,
    groups: Optional[Set[str]] = None,
) -> List[TestcaseValidationInfo]:
    def step():
        if progress is not None:
            progress.step()

    validation_entries = await extract_generation_testcases_from_groups(groups)
    validator_to_compiled_digest = compile_validators_for_entries(
        validation_entries, progress=progress
    )

    if not validator_to_compiled_digest:
        console.console.print(
            '[warning]No validators found, skipping validation.[/warning]'
        )
        if progress is not None:
            progress.omit()
        return []

    validation_info = []

    for entry in validation_entries:
        input_path = entry.metadata.copied_to.inputPath
        if not input_path.is_file():
            continue

        # Main validation.
        if entry.validator is not None:
            compiled_digest = validator_to_compiled_digest[str(entry.validator.path)]
            ok, message, hit_bounds = await validate_file(
                input_path,
                entry.validator,
                compiled_digest,
                group=entry.group_entry.group,
            )
            validation_info.append(
                TestcaseValidationInfo(
                    validator=entry.validator,
                    testcase=entry.group_entry,
                    generation_metadata=entry.metadata,
                    path=input_path,
                    ok=ok,
                    hit_bounds=hit_bounds,
                    message=message,
                )
            )

        for extra_validator in entry.extra_validators:
            compiled_digest = validator_to_compiled_digest[str(extra_validator.path)]
            ok, message, hit_bounds = await validate_file(
                input_path,
                extra_validator,
                compiled_digest,
                group=entry.group_entry.group,
            )
            validation_info.append(
                TestcaseValidationInfo(
                    validator=extra_validator,
                    testcase=entry.group_entry,
                    generation_metadata=entry.metadata,
                    path=input_path,
                    ok=ok,
                    hit_bounds=hit_bounds,
                    message=message,
                )
            )

        step()

    return validation_info


async def validate_outputs_from_entries(
    entries: List[GenerationTestcaseEntry],
    progress: Optional[StatusProgress] = None,
) -> List[TestcaseValidationInfo]:
    def step():
        if progress is not None:
            progress.step()

    validator_to_compiled_digest = compile_output_validators_for_entries(
        entries, progress=progress
    )

    if not validator_to_compiled_digest:
        console.console.print(
            '[warning]No output validators found, skipping validation.[/warning]'
        )
        if progress is not None:
            progress.omit()
        return []

    validation_info = []
    for entry in entries:
        output_path = entry.metadata.copied_to.outputPath
        if output_path is None or not output_path.is_file():
            continue
        for output_validator in entry.output_validators:
            compiled_digest = validator_to_compiled_digest[str(output_validator.path)]
            ok, message, _ = await validate_file(
                output_path,
                output_validator,
                compiled_digest,
                group=entry.group_entry.group,
            )
            validation_info.append(
                TestcaseValidationInfo(
                    validator=output_validator,
                    testcase=entry.group_entry,
                    generation_metadata=entry.metadata,
                    path=output_path,
                    ok=ok,
                    hit_bounds={},
                    message=message,
                )
            )
            step()

    return validation_info


async def check_output_from_entries(
    entries: List[GenerationTestcaseEntry],
    progress: Optional[StatusProgress] = None,
) -> List[TestcaseValidationInfo]:
    def step():
        if progress is not None:
            progress.step()

    entries_to_check: List[GenerationTestcaseEntry] = []

    for entry in entries:
        if entry.metadata.copied_from is None:
            continue
        if (
            entry.metadata.copied_from.outputPath is None
            or not entry.metadata.copied_from.outputPath.is_file()
        ):
            continue
        # Output is generated from a manual file, check it.
        entries_to_check.append(entry)

    if not entries_to_check:
        if progress is not None:
            progress.omit()
        return []

    checker = package.get_checker()
    checker_digest = checkers.compile_checker(progress=progress)

    validation_info = []
    for entry in entries_to_check:
        assert entry.metadata.copied_from is not None
        assert entry.metadata.copied_from.outputPath is not None
        assert entry.metadata.copied_from.outputPath.is_file()
        output_path = entry.metadata.copied_from.outputPath

        result = await checkers.check(
            checker_digest,
            run_log=None,
            testcase=entry.metadata.copied_to,
            program_output=output_path,
            skip_run_log=True,
        )

        if result.outcome != Outcome.ACCEPTED:
            validation_info.append(
                TestcaseValidationInfo(
                    validator=checker,
                    testcase=entry.group_entry,
                    generation_metadata=entry.metadata,
                    path=output_path,
                    ok=False,
                    hit_bounds={},
                    message=f'Checker failed on manual output: {result.message}',
                )
            )

        step()

    return validation_info


def has_validation_errors(infos: List[TestcaseValidationInfo]) -> bool:
    return any(not info.ok for info in infos)


def print_validation_report(
    infos: List[TestcaseValidationInfo], output_validation: bool = False
):
    any_failure = any(not info.ok for info in infos)
    validator_mode_str = 'output validator' if output_validation else 'validator'

    if output_validation and not any_failure:
        return

    if output_validation:
        console.console.rule('Output validation report', style='status')
    else:
        console.console.rule('Validation report', style='status')
    hit_bounds_per_group: Dict[Optional[str], HitBounds] = {}
    for info in infos:
        if not info.ok:
            console.console.print(
                f'[error]Testcase {info.href()} failed verification on {validator_mode_str} {info.validator.href()}:[/error]'
            )
            if info.generation_metadata is not None:
                metadata_markup = get_generation_metadata_markup(
                    info.generation_metadata
                )
                console.console.print(metadata_markup)
            console.console.print(info.message)
            continue

        if info.testcase is None:
            continue
        if info.testcase.group not in hit_bounds_per_group:
            hit_bounds_per_group[info.testcase.group] = {}
        hit_bounds_per_group[info.testcase.group] = _merge_hit_bounds(
            [hit_bounds_per_group[info.testcase.group], info.hit_bounds]
        )

    if not hit_bounds_per_group or output_validation:
        console.console.print()
        return

    if not _has_group_specific_validator():
        hit_bounds_per_group = {None: _merge_hit_bounds(hit_bounds_per_group.values())}

    def _is_hit_bound_good(hit_bounds: HitBounds) -> bool:
        return any(not v[0] or not v[1] for v in hit_bounds.values())

    # Cleanup entries in hit bounds per group that are totally empty.
    # Also skip samples.
    hit_bounds_per_group = {
        k: v
        for k, v in hit_bounds_per_group.items()
        if _is_hit_bound_good(v) and k != 'samples'
    }

    all_groups = set(info.testcase.group for info in infos if info.testcase is not None)
    if len(all_groups) == 1 and 'samples' in all_groups:
        # If there's only the samples group, do not check for hit bounds.
        hit_bounds_per_group = {}

    if not hit_bounds_per_group and not any_failure:
        console.console.print('No validation issues found.')
        return

    for group, hit_bounds in hit_bounds_per_group.items():
        if group is None:
            console.console.print('Hit bounds:')
        else:
            console.console.print(f'Group [item]{group}[/item] hit bounds:')

        for k, v in hit_bounds.items():
            if all(v):
                continue

            if not v[0]:
                console.console.print(
                    f'  - {k}: [warning]min-value not hit[/warning]',
                )
            if not v[1]:
                console.console.print(
                    f'  - {k}: [warning]max-value not hit[/warning]',
                )
        console.console.print()



import pathlib
import signal
from typing import Dict, Iterator, Tuple

import pytest

from rbx.box import checkers
from rbx.box.checkers import compile_checker
from rbx.box.schema import Testcase
from rbx.box.testing import testing_package
from rbx.grading.judge.sandbox import SandboxBase
from rbx.grading.limits import Limits
from rbx.grading.steps import Outcome, RunLog, RunLogMetadata

INTERESTING_CHECKERS = [
    'checkers/checker.cpp',
    'checkers/checker-fl.cpp',
    'checkers/checker-crash.cpp',
]


@pytest.fixture(scope='module')
def pkg_with_compiled_checker(tmp_path_factory):
    pkg_dir = tmp_path_factory.mktemp('pkg')
    with testing_package.TestingPackage(pkg_dir.absolute()) as testing_pkg:
        checkers = {}
        for checker in INTERESTING_CHECKERS:
            testing_pkg.set_checker('checker.cpp', src=checker)
            checkers[checker] = compile_checker()
        yield testing_pkg, checkers


@pytest.fixture(autouse=True)
def checker_pkg(
    pkg_with_compiled_checker: Tuple[testing_package.TestingPackage, str],
    testing_pkg: testing_package.TestingPackage,
) -> Iterator[testing_package.TestingPackage]:
    pkg, _ = pkg_with_compiled_checker
    testing_pkg.copy_from(pkg)
    with testing_pkg:
        yield testing_pkg


@pytest.fixture
def checker_digest(
    pkg_with_compiled_checker: Tuple[testing_package.TestingPackage, Dict[str, str]],
) -> str:
    _, checkers = pkg_with_compiled_checker
    return checkers['checkers/checker.cpp']


@pytest.fixture
def checker_digest_dict(
    pkg_with_compiled_checker: Tuple[testing_package.TestingPackage, Dict[str, str]],
) -> Dict[str, str]:
    _, checkers = pkg_with_compiled_checker
    return checkers


@pytest.fixture
def testcase(tmp_path_factory) -> Testcase:
    testcase_dir = tmp_path_factory.mktemp('testcase')
    input_path = testcase_dir / 'input.txt'
    input_path.touch()
    output_path = testcase_dir / 'output.txt'
    output_path.touch()
    return Testcase(
        inputPath=input_path,
        outputPath=output_path,
    )


@pytest.fixture
def program_output(tmp_path_factory) -> pathlib.Path:
    output_dir = tmp_path_factory.mktemp('output')
    output_path = output_dir / 'output.txt'
    output_path.touch()
    return output_path


@pytest.fixture
def interactor_stderr(tmp_path_factory) -> pathlib.Path:
    stderr_dir = tmp_path_factory.mktemp('stderr')
    stderr_path = stderr_dir / 'stderr.txt'
    stderr_path.touch()
    return stderr_path


@pytest.fixture
def testlib_eof_stderr(interactor_stderr: pathlib.Path) -> pathlib.Path:
    interactor_stderr.write_text('wrong output format Unexpected end of file')
    return interactor_stderr


@pytest.fixture
def run_log() -> RunLog:
    return RunLog(
        exitcode=0,
        exitstatus=SandboxBase.EXIT_OK,
        metadata=RunLogMetadata(
            limits=Limits(time=1000, memory=1024, output=1024, isDoubleTL=True),
            timeLimit=1000,
            memoryLimit=1024,
            retryIndex=0,
        ),
        exitindex=0,
    )


@pytest.fixture
def run_log_with_warnings(run_log: RunLog) -> RunLog:
    run_log.warnings = True
    return run_log


@pytest.fixture
def interactor_run_log() -> RunLog:
    return RunLog(
        exitcode=0,
        exitstatus=SandboxBase.EXIT_OK,
        metadata=RunLogMetadata(
            limits=Limits(time=1000, memory=1024, output=1024, isDoubleTL=True),
            timeLimit=1000,
            memoryLimit=1024,
            retryIndex=0,
        ),
        exitindex=1,
    )


@pytest.fixture
def interactor_exit_first(run_log: RunLog, interactor_run_log: RunLog):
    interactor_run_log.exitindex = 0
    run_log.exitindex = 1


class TestCheckCommunicationInternalError:
    async def test_check_communication_sol_run_log_is_none(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        result = await checkers.check_communication(
            checker_digest,
            None,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.INTERNAL_ERROR

    async def test_check_communication_interactor_run_log_is_none(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        result = await checkers.check_communication(
            checker_digest,
            run_log,
            None,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.INTERNAL_ERROR

    async def test_check_communication_sol_run_log_has_sandbox_error(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitstatus = SandboxBase.EXIT_SANDBOX_ERROR
        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.INTERNAL_ERROR

    async def test_check_communication_interactor_run_log_has_sandbox_error(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        interactor_run_log.exitstatus = SandboxBase.EXIT_SANDBOX_ERROR
        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.INTERNAL_ERROR


@pytest.mark.usefixtures('interactor_exit_first')
class TestCheckCommunicationInteractorExitedFirst:
    async def test_check_communication_wa(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN
        interactor_run_log.exitcode = 1
        interactor_run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.WRONG_ANSWER

    async def test_check_communication_sol_terminated_and_fl(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitstatus = SandboxBase.EXIT_TERMINATED
        interactor_run_log.exitcode = 3
        interactor_run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.JUDGE_FAILED

    async def test_check_communication_sol_terminated_and_interactor_rte(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitstatus = SandboxBase.EXIT_TERMINATED
        interactor_run_log.exitcode = 42
        interactor_run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.JUDGE_FAILED

    async def test_check_communication_sol_terminated_and_interactor_tle(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitstatus = SandboxBase.EXIT_TERMINATED
        interactor_run_log.exitstatus = SandboxBase.EXIT_TIMEOUT

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.TIME_LIMIT_EXCEEDED

    async def test_check_communication_sol_terminated_and_interactor_internal_error(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitstatus = SandboxBase.EXIT_TERMINATED
        interactor_run_log.exitstatus = SandboxBase.EXIT_SANDBOX_ERROR

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.INTERNAL_ERROR

    async def test_check_communication_sol_sigpipe_and_wa(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitcode = -signal.SIGPIPE
        interactor_run_log.exitcode = 1
        interactor_run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.WRONG_ANSWER

    async def test_check_communication_sol_sigpipe_as_rte_and_wa(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitcode = 2
        run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN
        interactor_run_log.exitcode = 1
        interactor_run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.WRONG_ANSWER


class TestCheckCommunicatedSolutionCrashedFirst:
    async def test_check_communication_sol_rte(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        testlib_eof_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitcode = 2
        run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            testlib_eof_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.RUNTIME_ERROR

    async def test_check_communication_sol_mle(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        testlib_eof_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitstatus = SandboxBase.EXIT_MEMORY_LIMIT_EXCEEDED

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            testlib_eof_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.MEMORY_LIMIT_EXCEEDED

    async def test_check_communication_sol_tle(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        testlib_eof_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitstatus = SandboxBase.EXIT_TIMEOUT

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            testlib_eof_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.TIME_LIMIT_EXCEEDED


class TestCheckCommunicatedSolutionExitedOkFirst:
    async def test_check_communication_sol_ok_interactor_wa(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        interactor_run_log.exitcode = 1
        interactor_run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.WRONG_ANSWER

    async def test_check_communication_sol_ok_interactor_fl(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        interactor_run_log.exitcode = 3
        interactor_run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.JUDGE_FAILED

    async def test_check_communication_sol_ok_interactor_rte(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        interactor_run_log.exitcode = 42
        interactor_run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.JUDGE_FAILED

    async def test_check_communication_sol_ok_interactor_tle(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        interactor_run_log.exitstatus = SandboxBase.EXIT_TIMEOUT

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.TIME_LIMIT_EXCEEDED


class TestCheckCommunicationLegacyChecker:
    async def test_check_communication_checker_ac(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        assert testcase.outputPath is not None
        testcase.outputPath.write_text('123\n')
        program_output.write_text('123\n')

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.ACCEPTED

    async def test_check_communication_checker_wa(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        assert testcase.outputPath is not None
        testcase.outputPath.write_text('123\n')
        program_output.write_text('456\n')

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.WRONG_ANSWER


class TestCheckCommunicationNoChecker:
    """Test scenarios when no checker is provided (checker_digest is None)."""

    async def test_check_communication_no_checker_solution_accepted(
        self,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        result = await checkers.check_communication(
            None,  # No checker
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.ACCEPTED

    async def test_check_communication_no_checker_solution_rte(
        self,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            None,  # No checker
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.RUNTIME_ERROR

    async def test_check_communication_no_checker_interactor_wa(
        self,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        interactor_run_log.exitcode = 1
        interactor_run_log.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            None,  # No checker
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.WRONG_ANSWER


class TestCheckCommunicationSanitizerWarnings:
    """Test sanitizer warnings propagation in communication mode."""

    async def test_check_communication_sanitizer_warnings_on_accepted(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log_with_warnings: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        assert testcase.outputPath is not None
        testcase.outputPath.write_text('123\n')
        program_output.write_text('123\n')

        result = await checkers.check_communication(
            checker_digest,
            run_log_with_warnings,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.ACCEPTED
        assert result.sanitizer_warnings is True

    async def test_check_communication_sanitizer_warnings_on_failure(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log_with_warnings: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        run_log_with_warnings.exitstatus = SandboxBase.EXIT_NONZERO_RETURN

        result = await checkers.check_communication(
            checker_digest,
            run_log_with_warnings,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.RUNTIME_ERROR
        assert result.sanitizer_warnings is True

    async def test_check_communication_no_sanitizer_warnings(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        assert testcase.outputPath is not None
        testcase.outputPath.write_text('123\n')
        program_output.write_text('123\n')

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.ACCEPTED
        assert result.sanitizer_warnings is False


class TestCheckCommunicationSkipRunLog:
    """Test scenarios with skip_run_log=True."""

    async def test_check_communication_skip_run_log_solution_accepted(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        assert testcase.outputPath is not None
        testcase.outputPath.write_text('123\n')
        program_output.write_text('123\n')

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
            skip_run_log=True,
        )

        assert result.outcome == Outcome.ACCEPTED

    async def test_check_communication_skip_run_log_with_bad_solution(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        assert testcase.outputPath is not None
        testcase.outputPath.write_text('123\n')
        program_output.write_text('456\n')  # Wrong answer

        # skip_run_log only affects the final checker run, not the pre-checks
        # So we need a scenario where the solution is OK but the checker would fail
        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
            skip_run_log=True,
        )

        # The checker should still run and detect wrong answer
        assert result.outcome == Outcome.WRONG_ANSWER


class TestCheckCommunicationInteractorComplexScenarios:
    """Test complex interactor scenarios and edge cases."""

    async def test_check_communication_interactor_mle(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        interactor_run_log.exitstatus = SandboxBase.EXIT_MEMORY_LIMIT_EXCEEDED

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.MEMORY_LIMIT_EXCEEDED

    async def test_check_communication_interactor_output_limit(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        interactor_run_log.exitstatus = SandboxBase.EXIT_OUTPUT_LIMIT_EXCEEDED

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.OUTPUT_LIMIT_EXCEEDED

    async def test_check_communication_interactor_idleness(
        self,
        checker_digest: str,
        testcase: Testcase,
        program_output: pathlib.Path,
        interactor_stderr: pathlib.Path,
        run_log: RunLog,
        interactor_run_log: RunLog,
    ) -> None:
        interactor_run_log.exitstatus = SandboxBase.EXIT_TIMEOUT_WALL

        result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_stderr,
            testcase,
            program_output,
        )

        assert result.outcome == Outcome.IDLENESS_LIMIT_EXCEEDED



"""FastAPI application for robox.io webserver."""

from pathlib import Path
from typing import List, Optional

from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, RedirectResponse
from pydantic import BaseModel

from rbx import console
from rbx.box import package
from rbx.box.environment import VerificationLevel
from rbx.box.schema import CodeItem
from rbx.box.statements import build_statements
from rbx.box.statements.schema import Statement, StatementType
from rbx.box.wizard.llm_router import router as llm_router

origins = [
    'http://localhost',
    'http://localhost:3000',  # Replace with your client-side port
    'http://localhost:3001',  # Replace with your client-side port
]

# Create FastAPI instance
app = FastAPI(title='Robox.io Wizard', version='0.1.0')
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)


app.include_router(llm_router)


@app.get('/api/')
async def hello_world():
    """Return a Hello World JSON response."""
    return {'message': 'Hello World', 'status': 'success'}


class CodeResponse(BaseModel):
    path: str
    code: str
    language: Optional[str]


class CodeUpdateRequest(BaseModel):
    code: str


def _serve_code(code: CodeItem):
    if not code.path.is_file():
        raise HTTPException(status_code=404, detail='Code file not found')
    return CodeResponse(
        path=str(code.path), code=code.path.read_text(), language=code.language
    )


@app.get('/checker')
async def serve_checker():
    """Serve the checker file."""
    pkg = package.find_problem_package_or_die()
    if pkg.checker is None:
        raise HTTPException(status_code=404, detail='Checker not found')
    return _serve_code(pkg.checker)


@app.put('/checker')
async def update_checker(payload: CodeUpdateRequest):
    """Update the checker file."""
    pkg = package.find_problem_package_or_die()
    if pkg.checker is None:
        raise HTTPException(status_code=404, detail='Checker not found')
    pkg.checker.path.write_text(payload.code)
    return _serve_code(pkg.checker)


@app.get('/interactor')
async def serve_interactor():
    """Serve the checker file."""
    pkg = package.find_problem_package_or_die()
    if pkg.interactor is None:
        raise HTTPException(status_code=404, detail='Interactor not found')
    return _serve_code(pkg.interactor)


@app.put('/interactor')
async def update_interactor(payload: CodeUpdateRequest):
    """Update the interactor file."""
    pkg = package.find_problem_package_or_die()
    if pkg.interactor is None:
        raise HTTPException(status_code=404, detail='Interactor not found')
    pkg.interactor.path.write_text(payload.code)
    return _serve_code(pkg.interactor)


@app.get('/validator')
async def serve_validator():
    """Serve the validator file."""
    pkg = package.find_problem_package_or_die()
    if pkg.validator is None:
        raise HTTPException(status_code=404, detail='Validator not found')
    return _serve_code(pkg.validator)


@app.put('/validator')
async def update_validator(payload: CodeUpdateRequest):
    """Update the validator file."""
    pkg = package.find_problem_package_or_die()
    if pkg.validator is None:
        raise HTTPException(status_code=404, detail='Validator not found')
    pkg.validator.path.write_text(payload.code)
    return _serve_code(pkg.validator)


@app.get('/statement/{statement_name}')
async def serve_statement(statement_name: str):
    """Serve the statement file."""
    statement = package.get_statement_or_nil(statement_name)
    if statement is None:
        raise HTTPException(status_code=404, detail='Statement not found')
    build_path = build_statements.get_statement_build_path(statement, StatementType.PDF)
    return RedirectResponse(f'/package/{build_path}')


@app.get('/statement')
async def serve_main_statement():
    """Serve the main statement file."""
    pkg = package.find_problem_package_or_die()
    if not pkg.expanded_statements:
        raise HTTPException(status_code=404, detail='Statement not found')
    statement = pkg.expanded_statements[0]
    return RedirectResponse(f'/statement/{statement.name}')


@app.post('/statement/{statement_name}/build')
async def build_statement(statement_name: str):
    """Build the statement file."""
    statement = package.get_statement_or_nil(statement_name)
    if statement is None:
        raise HTTPException(status_code=404, detail='Statement not found')
    try:
        with console.console.capture() as capture:
            await build_statements.execute_build(
                verification=VerificationLevel.VALIDATE.value, names=[statement_name]
            )
    except Exception as e:
        raise HTTPException(status_code=500, detail=capture.get() + str(e)) from None


@app.get('/statement/{statement_name}/code')
async def serve_statement_code(statement_name: str):
    """Serve the statement code file."""
    statement = package.get_statement_or_nil(statement_name)
    if statement is None:
        raise HTTPException(status_code=404, detail='Statement not found')
    if statement.type != StatementType.rbxTeX:
        raise HTTPException(
            status_code=400, detail='Statement is not a rbxTeX statement'
        )
    return _serve_code(CodeItem(path=statement.path, language='latex'))


@app.put('/statement/{statement_name}/code')
async def update_statement_code(statement_name: str, payload: CodeUpdateRequest):
    """Update the statement code file."""
    statement = package.get_statement_or_nil(statement_name)
    if statement is None:
        raise HTTPException(status_code=404, detail='Statement not found')
    if statement.type != StatementType.rbxTeX:
        raise HTTPException(
            status_code=400, detail='Statement is not a rbxTeX statement'
        )
    statement.path.write_text(payload.code)
    return _serve_code(CodeItem(path=statement.path, language='latex'))


class StatementListResponse(BaseModel):
    statements: List[Statement]


@app.get('/statements')
async def serve_statements():
    """Serve the statements file."""
    pkg = package.find_problem_package_or_die()
    return StatementListResponse(statements=pkg.expanded_statements)


@app.get('/package/{path:path}')
async def serve_package_file(path: str):
    """Serve files from the current working directory.

    Args:
        path: The file path relative to the current working directory.

    Returns:
        FileResponse: The requested file.

    Raises:
        HTTPException: If the file doesn't exist or is outside the working directory.
    """
    # Get the current working directory
    cwd = Path.cwd()

    # Resolve the full path
    file_path = cwd / path

    # Resolve to absolute path to handle .. and symbolic links
    try:
        file_path = file_path.resolve()
    except Exception:
        raise HTTPException(status_code=404, detail='Invalid file path') from None

    # Security check: ensure the resolved path is within the current working directory
    try:
        file_path.relative_to(cwd)
    except ValueError:
        raise HTTPException(
            status_code=403,
            detail='Access denied: Path is outside the working directory',
        ) from None

    # Check if the file exists and is a file (not a directory)
    if not file_path.exists():
        raise HTTPException(status_code=404, detail='File not found')

    if not file_path.is_file():
        raise HTTPException(status_code=404, detail='Path is not a file')

    # Serve the file
    return FileResponse(file_path)



import math
import pathlib
from functools import reduce
from typing import Any, Callable, Optional, Union

import simpleeval

NameNotDefined = simpleeval.NameNotDefined
AttributeDoesNotExist = simpleeval.AttributeDoesNotExist

PathLike = Union[str, pathlib.Path]


def _step_down(x: Any, step: int) -> int:
    x = int(x)
    return x // step * step


def _step_up(x: Any, step: int) -> int:
    x = int(x)
    return (x + step - 1) // step * step


def _step_closest(x: Any, step: int) -> int:
    x = int(x)
    return round(x / step) * step


def _path_ext(path: pathlib.Path) -> str:
    return path.suffix[1:]


def _with_path_ext(path: pathlib.Path, ext: str) -> pathlib.Path:
    return path.with_suffix(f'.{ext}')


def _path_fn(fn: Callable[[pathlib.Path], PathLike]) -> Callable[[PathLike], str]:
    def res_fn(path: PathLike) -> str:
        return str(fn(pathlib.Path(path)))

    return res_fn


def _path_fn_2args(
    fn: Callable[[pathlib.Path, str], PathLike],
) -> Callable[[PathLike, str], str]:
    def res_fn(path: PathLike, arg: str) -> str:
        return str(fn(pathlib.Path(path), arg))

    return res_fn


def _path_stem(path: pathlib.Path) -> str:
    return path.stem


def _path_parent(path: pathlib.Path) -> pathlib.Path:
    return path.parent


def _path_suffix(path: pathlib.Path) -> str:
    return path.suffix


def _get_functions(functions: Optional[dict[str, Any]]) -> dict[str, Any]:
    res = {}
    res.update(
        # Math functions.
        int=int,
        float=float,
        str=str,
        bool=bool,
        len=len,
        floor=math.floor,
        ceil=math.ceil,
        round=round,
        abs=abs,
        step_down=_step_down,
        step_up=_step_up,
        step_closest=_step_closest,
        max=max,
        min=min,
        sum=sum,
        map=map,
        zip=zip,
        filter=filter,
        reduce=reduce,
        # Path functions.
        stem=_path_fn(_path_stem),
        parent=_path_fn(_path_parent),
        suffix=_path_fn(_path_suffix),
        ext=_path_fn(_path_ext),
        with_suffix=_path_fn_2args(pathlib.Path.with_suffix),
        with_stem=_path_fn_2args(pathlib.Path.with_stem),
        with_ext=_path_fn_2args(_with_path_ext),
    )
    res.update(functions or {})
    return res


def eval(
    expression: str,
    names: Optional[dict[str, Any]] = None,
    functions: Optional[dict[str, Any]] = None,
) -> Any:
    return simpleeval.simple_eval(
        expression, names=names, functions=_get_functions(functions)
    )


def eval_int(
    expression: str,
    names: Optional[dict[str, Any]] = None,
    functions: Optional[dict[str, Any]] = None,
) -> int:
    return int(eval(expression, names, functions))


def eval_float(
    expression: str,
    names: Optional[dict[str, Any]] = None,
    functions: Optional[dict[str, Any]] = None,
) -> float:
    return float(eval(expression, names, functions))


def eval_string(
    expression: str,
    names: Optional[dict[str, Any]] = None,
    functions: Optional[dict[str, Any]] = None,
) -> str:
    return f'{eval(expression, names, functions)}'


def eval_as_fstring(
    expression: str,
    names: Optional[dict[str, Any]] = None,
    functions: Optional[dict[str, Any]] = None,
) -> str:
    # Escape single quotes in the expression so f-string is
    # parsed correctly.
    expression = expression.replace("'", "\\'")
    return eval_string(f"f'{expression}'", names, functions)



"""
https://github.com/citxx/polygon-py

Manually copied because can't add direct dependency with Poetry.
"""
# flake8: noqa

import hashlib
import json
import random
import string
import time
from enum import Enum

import requests
import typer

from rbx import console


class Polygon:
    """ """

    # API methods
    _CONTEST_PROBLEMS = 'contest.problems'
    _PROBLEM_INFO = 'problem.info'
    _PROBLEM_UPDATE_INFO = 'problem.updateInfo'
    _PROBLEM_UPDATE_WORKING_COPY = 'problem.updateWorkingCopy'
    _PROBLEM_COMMIT_CHANGES = 'problem.commitChanges'
    _PROBLEM_STATEMENTS = 'problem.statements'
    _PROBLEM_SAVE_STATEMENT = 'problem.saveStatement'
    _PROBLEM_STATEMENT_RESOURCES = 'problem.statementResources'
    _PROBLEM_SAVE_STATEMENT_RESOURCE = 'problem.saveStatementResource'
    _PROBLEMS_LIST = 'problems.list'
    _PROBLEM_CREATE = 'problem.create'
    _PROBLEM_CHECKER = 'problem.checker'
    _PROBLEM_VALIDATOR = 'problem.validator'
    _PROBLEM_INTERACTOR = 'problem.interactor'
    _PROBLEM_FILES = 'problem.files'
    _PROBLEM_SOLUTIONS = 'problem.solutions'
    _PROBLEM_VIEW_FILE = 'problem.viewFile'
    _PROBLEM_VIEW_SOLUTION = 'problem.viewSolution'
    _PROBLEM_SCRIPT = 'problem.script'
    _PROBLEM_TESTS = 'problem.tests'
    _PROBLEM_TEST_INPUT = 'problem.testInput'
    _PROBLEM_TEST_ANSWER = 'problem.testAnswer'
    _PROBLEM_SET_VALIDATOR = 'problem.setValidator'
    _PROBLEM_SET_CHECKER = 'problem.setChecker'
    _PROBLEM_SET_INTERACTOR = 'problem.setInteractor'
    _PROBLEM_SAVE_FILE = 'problem.saveFile'
    _PROBLEM_SAVE_SOLUTION = 'problem.saveSolution'
    _PROBLEM_EDIT_SOLUTION_EXTRA_TAGS = 'problem.editSolutionExtraTags'
    _PROBLEM_SAVE_SCRIPT = 'problem.saveScript'
    _PROBLEM_SAVE_TEST = 'problem.saveTest'
    _PROBLEM_ENABLE_GROUPS = 'problem.enableGroups'
    _PROBLEM_ENABLE_POINTS = 'problem.enablePoints'
    _PROBLEM_VIEW_TEST_GROUP = 'problem.viewTestGroup'
    _PROBLEM_SAVE_TEST_GROUP = 'problem.saveTestGroup'
    _PROBLEM_VIEW_TAGS = 'problem.viewTags'
    _PROBLEM_SAVE_TAGS = 'problem.saveTags'
    _PROBLEM_VIEW_GENERAL_DESCRIPTION = 'problem.viewGeneralDescription'
    _PROBLEM_SAVE_GENERAL_DESCRIPTION = 'problem.saveGeneralDescription'
    _PROBLEM_VIEW_GENERAL_TUTORIAL = 'problem.viewGeneralTutorial'
    _PROBLEM_SAVE_GENERAL_TUTORIAL = 'problem.saveGeneralTutorial'

    def __init__(self, api_url, api_key, api_secret):
        self.request_config = RequestConfig(api_url, api_key, api_secret)

    def problems_list(self, show_deleted=None, id=None, name=None, owner=None):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEMS_LIST,
            args={
                'showDeleted': show_deleted,
                'id': id,
                'name': name,
                'owner': owner,
            },
        )
        return [Problem.from_json(self, p_json) for p_json in response.result]

    def problem_create(self, name):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_CREATE,
            args={
                'name': name,
            },
        )
        return Problem.from_json(self, response.result)

    def problem_info(self, problem_id):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_INFO,
            args={'problemId': problem_id},
        )
        return ProblemInfo.from_json(response.result)

    def problem_update_info(self, problem_id, problem_info):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_UPDATE_INFO,
            args={
                'problemId': problem_id,
                'inputFile': problem_info.input_file,
                'outputFile': problem_info.output_file,
                'interactive': problem_info.interactive,  # TODO
                'timeLimit': problem_info.time_limit,
                'memoryLimit': problem_info.memory_limit,
            },
        )
        return response.result

    def problem_update_working_copy(self, problem_id):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_UPDATE_WORKING_COPY,
            args={'problemId': problem_id},
        )
        return response.result

    def problem_commit_changes(self, problem_id, minor_changes=None, message=None):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_COMMIT_CHANGES,
            args={
                'problemId': problem_id,
                'minorChanges': minor_changes,
                'message': message,
            },
        )
        return response.result

    def problem_view_tags(self, problem_id):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_VIEW_TAGS,
            args={'problemId': problem_id},
        )
        return response.result

    def problem_save_tags(self, problem_id, tags):
        """ """
        tags_str = ','.join(tags)
        response = self._request_ok_or_raise(
            self._PROBLEM_SAVE_TAGS,
            args={
                'problemId': problem_id,
                'tags': tags_str,
            },
        )
        return response.result

    def problem_view_general_description(self, problem_id):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_VIEW_GENERAL_DESCRIPTION,
            args={'problemId': problem_id},
        )
        return response.result

    def problem_save_general_description(self, problem_id, description):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_SAVE_GENERAL_DESCRIPTION,
            args={
                'problemId': problem_id,
                'description': description,
            },
        )
        return response.result

    def problem_view_general_tutorial(self, problem_id):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_VIEW_GENERAL_TUTORIAL,
            args={'problemId': problem_id},
        )
        return response.result

    def problem_save_general_tutorial(self, problem_id, tutorial):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_SAVE_GENERAL_TUTORIAL,
            args={
                'problemId': problem_id,
                'tutorial': tutorial,
            },
        )
        return response.result

    def problem_statements(self, problem_id):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_STATEMENTS,
            args={
                'problemId': problem_id,
            },
        )
        return {
            lang: Statement.from_json(statement_json)
            for lang, statement_json in response.result.items()
        }

    def problem_save_statement(self, problem_id, lang, problem_statement):
        """ """
        if not isinstance(problem_statement, Statement):
            raise ValueError(
                'Expected Statement instance for problem_statement argument, but %s found'
                % type(problem_statement)
            )
        response = self._request_ok_or_raise(
            self._PROBLEM_SAVE_STATEMENT,
            args={
                'problemId': problem_id,
                'lang': lang,
                'encoding': problem_statement.encoding,
                'name': problem_statement.name,
                'legend': problem_statement.legend,
                'input': problem_statement.input,
                'output': problem_statement.output,
                'scoring': problem_statement.scoring,
                'interaction': problem_statement.interaction,
                'notes': problem_statement.notes,
                'tutorial': problem_statement.tutorial,
            },
        )
        return response.result

    def problem_statement_resources(self, problem_id):
        response = self._request_ok_or_raise(
            self._PROBLEM_STATEMENT_RESOURCES,
            args={
                'problemId': problem_id,
            },
        )
        return response.result

    def problem_save_statement_resource(
        self, problem_id, name, file, check_existing=None
    ):
        response = self._request_ok_or_raise(
            self._PROBLEM_SAVE_STATEMENT_RESOURCE,
            args={
                'problemId': problem_id,
                'name': name,
                'file': file,
                'checkExisting': check_existing,
            },
        )
        return response.result

    def problem_enable_groups(self, problem_id, testset, enable):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_ENABLE_GROUPS,
            args={
                'problemId': problem_id,
                'testset': testset,
                'enable': enable,
            },
        )
        return response.result

    def problem_enable_points(self, problem_id, enable):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_ENABLE_POINTS,
            args={
                'problemId': problem_id,
                'enable': enable,
            },
        )
        return response.result

    def problem_save_script(self, problem_id, testset, source):
        response = self._request_ok_or_raise(
            self._PROBLEM_SAVE_SCRIPT,
            args={
                'problemId': problem_id,
                'testset': testset,
                'source': source,
            },
        )
        return response.result

    def problem_save_test(
        self,
        problem_id,
        testset,
        test_index,
        test_input,
        test_group=None,
        test_points=None,
        test_description=None,
        test_use_in_statements=None,
        test_input_for_statements=None,
        test_output_for_statements=None,
        verify_input_output_for_statements=None,
        check_existing=None,
    ):
        """ """
        response = self._request_ok_or_raise(
            self._PROBLEM_SAVE_TEST,
            args={
                'problemId': problem_id,
                'testset': testset,
                'testIndex': test_index,
                'testInput': test_input,
                'testGroup': test_group,
                'testPoints': test_points,
                'testDescription': test_description,
                'testUseInStatements': test_use_in_statements,
                'testInputForStatements': test_input_for_statements,
                'testOutputForStatements': test_output_for_statements,
                'verifyInputOutputForStatements': verify_input_output_for_statements,
                'checkExisting': check_existing,
            },
        )
        return response.result

    def problem_solutions(self, problem_id):
        response = self._request_ok_or_raise(
            self._PROBLEM_SOLUTIONS,
            args={
                'problemId': problem_id,
            },
        )
        return [Solution.from_json(js) for js in response.result]

    def problem_files(self, problem_id):
        response = self._request_ok_or_raise(
            self._PROBLEM_FILES,
            args={
                'problemId': problem_id,
            },
        )
        return {
            type: [File.from_json(js) for js in response.result[str(type) + 'Files']]
            for type in FileType
        }

    def problem_tests(self, problem_id, testset):
        response = self._request_ok_or_raise(
            self._PROBLEM_TESTS,
            args={
                'problemId': problem_id,
                'testset': testset,
            },
        )
        return [Test.from_json(self, problem_id, testset, js) for js in response.result]

    def problem_save_test_group(
        self,
        problem_id,
        testset,
        group,
        points_policy=None,
        feedback_policy=None,
        dependencies=None,
    ):
        if isinstance(dependencies, list):
            dependencies = ','.join(map(str, dependencies))
        elif dependencies is not None:
            dependencies = str(dependencies)
        if points_policy is not None and not isinstance(points_policy, PointsPolicy):
            raise ValueError(
                'Expected PointsPolicy instance for points_policy argument, but %s found'
                % type(points_policy)
            )
        if feedback_policy is not None and not isinstance(
            feedback_policy, FeedbackPolicy
        ):
            raise ValueError(
                'Expected FeedbackPolicy instance for feedback_policy argument, but %s found'
                % type(feedback_policy)
            )
        response = self._request_ok_or_raise(
            self._PROBLEM_SAVE_TEST_GROUP,
            args={
                'problemId': problem_id,
                'testset': testset,
                'group': group,
                'pointsPolicy': points_policy,
                'feedbackPolicy': feedback_policy,
                'dependencies': dependencies,
            },
        )
        return response.result

    def problem_view_test_group(self, testset, group):
        response = self._request_ok_or_raise(
            self._PROBLEM_VIEW_TEST_GROUP,
            args={
                'testset': testset,
                'group': group,
            },
        )
        return TestGroup.from_json(response.result)

    def problem_view_file(self, problem_id, type, name):
        response = self._request_raw(
            self._PROBLEM_VIEW_FILE,
            args={'problemId': problem_id, 'type': type, 'name': name},
        )
        return response

    def problem_save_file(
        self,
        problem_id,
        type,
        name,
        file,
        source_type=None,
        resource_advanced_properties=None,
    ):
        stages = (
            None
            if resource_advanced_properties is None
            or resource_advanced_properties.stages is None
            else ';'.join(map(str, resource_advanced_properties.stages))
        )
        assets = (
            None
            if resource_advanced_properties is None
            or resource_advanced_properties.assets is None
            else ';'.join(map(str, resource_advanced_properties.assets))
        )
        response = self._request_ok_or_raise(
            self._PROBLEM_SAVE_FILE,
            args={
                'problemId': problem_id,
                'type': type,
                'name': name,
                'file': file,
                'sourceType': source_type,
                'forTypes': None
                if resource_advanced_properties is None
                else resource_advanced_properties.for_types,
                'stages': stages,
                'assets': assets,
            },
        )
        return response.result

    def problem_save_solution(
        self, problem_id, name, file, source_type, tag, check_existing=None
    ):
        response = self._request_ok_or_raise(
            self._PROBLEM_SAVE_SOLUTION,
            args={
                'problemId': problem_id,
                'name': name,
                'file': file,
                'sourceType': source_type,
                'tag': tag,
                'checkExisting': check_existing,
            },
        )
        return response.result

    def problem_checker(self, problem_id):
        response = self._request_ok_or_raise(
            self._PROBLEM_CHECKER,
            args={
                'problemId': problem_id,
            },
        )
        return response.result

    def problem_set_checker(self, problem_id, checker):
        response = self._request_ok_or_raise(
            self._PROBLEM_SET_CHECKER,
            args={
                'problemId': problem_id,
                'checker': checker,
            },
        )
        return response.result

    def problem_validator(self, problem_id):
        response = self._request_ok_or_raise(
            self._PROBLEM_VALIDATOR,
            args={
                'problemId': problem_id,
            },
        )
        return response.result

    def problem_set_validator(self, problem_id, validator):
        response = self._request_ok_or_raise(
            self._PROBLEM_SET_VALIDATOR,
            args={
                'problemId': problem_id,
                'validator': validator,
            },
        )
        return response.result

    def problem_interactor(self, problem_id):
        response = self._request_ok_or_raise(
            self._PROBLEM_INTERACTOR,
            args={
                'problemId': problem_id,
            },
        )
        return response.result

    def problem_set_interactor(self, problem_id, interactor):
        response = self._request_ok_or_raise(
            self._PROBLEM_SET_INTERACTOR,
            args={
                'problemId': problem_id,
                'interactor': interactor,
            },
        )
        return response.result

    def contest_problems(self, contest_id):
        """ """
        response = self._request_ok_or_raise(
            self._CONTEST_PROBLEMS,
            args={'contestId': contest_id},
        )
        return {
            name: Problem.from_json(self, p_json)
            for name, p_json in response.result.items()
        }

    def _request_ok_or_raise(self, method_name, args=None):
        response = self._request(method_name, args=args)
        if response.status != Response.STATUS_OK:
            raise PolygonRequestFailedException(response.comment)
        return response

    def _request(self, method_name, args=None):
        request = Request(self.request_config, method_name, args)
        return request.issue()

    def _request_raw(self, method_name, args=None):
        request = Request(self.request_config, method_name, args)
        return request.issue_raw()


class Problem:
    """
    Object representing Polygon problem
    """

    # JSON field names
    _ID_FIELD = 'id'
    _OWNER_FIELD = 'owner'
    _NAME_FIELD = 'name'
    _DELETED_FIELD = 'deleted'
    _FAVORITE_FIELD = 'favourite'
    _ACCESS_TYPE_FIELD = 'accessType'
    _REVISION_FIELD = 'revision'
    _LATEST_PACKAGE_FIELD = 'latestPackage'
    _MODIFIED_FIELD = 'modified'

    @classmethod
    def from_json(cls, polygon, problem_json):
        return cls(
            polygon=polygon,
            problem_id=problem_json.get(Problem._ID_FIELD),
            owner=problem_json.get(Problem._OWNER_FIELD),
            name=problem_json.get(Problem._NAME_FIELD),
            deleted=problem_json.get(Problem._DELETED_FIELD),
            favorite=problem_json.get(Problem._FAVORITE_FIELD),
            access_type=problem_json.get(Problem._ACCESS_TYPE_FIELD),
            revision=problem_json.get(Problem._REVISION_FIELD),
            latest_package=problem_json.get(Problem._LATEST_PACKAGE_FIELD),
            modified=problem_json.get(Problem._MODIFIED_FIELD),
        )

    def __init__(
        self,
        polygon,
        problem_id,
        owner,
        name,
        deleted,
        favorite,
        access_type,
        revision,
        latest_package,
        modified,
    ):
        self._polygon = polygon

        self.id = problem_id
        self.owner = owner
        self.name = name
        self.deleted = deleted
        self.favorite = favorite
        self.access_type = access_type
        self.revision = revision
        self.latest_package = latest_package
        self.modified = modified

    def __str__(self):
        return '{}:{}'.format(self.name, self.id)

    def info(self):
        return self._polygon.problem_info(self.id)

    def update_info(self, problem_info):
        return self._polygon.problem_update_info(self.id, problem_info)

    def update_working_copy(self):
        return self._polygon.problem_update_working_copy(self.id)

    def commit_changes(self, minor_changes=None, message=None):
        return self._polygon.problem_commit_changes(self.id, minor_changes, message)

    def tags(self):
        return self._polygon.problem_view_tags(self.id)

    def save_tags(self, tags):
        return self._polygon.problem_save_tags(self.id, tags)

    def general_description(self):
        return self._polygon.problem_view_general_description(self.id)

    def save_general_description(self, description):
        return self._polygon.problem_save_general_description(self.id, description)

    def general_tutorial(self):
        return self._polygon.problem_view_general_tutorial(self.id)

    def save_general_tutorial(self, tutorial):
        return self._polygon.problem_save_general_tutorial(self.id, tutorial)

    def statements(self):
        return self._polygon.problem_statements(self.id)

    def save_statement(self, lang, problem_statement):
        return self._polygon.problem_save_statement(self.id, lang, problem_statement)

    def statement_resources(self):
        return self._polygon.problem_statement_resources(self.id)

    def save_statement_resource(self, name, file, check_existing=None):
        return self._polygon.problem_save_statement_resource(
            self.id, name, file, check_existing
        )

    def enable_groups(self, testset, enable):
        return self._polygon.problem_enable_groups(self.id, testset, enable)

    def enable_points(self, enable):
        return self._polygon.problem_enable_points(self.id, enable)

    def save_script(self, testset, source):
        return self._polygon.problem_save_script(self.id, testset, source)

    def save_test(
        self,
        testset,
        test_index,
        test_input,
        test_group=None,
        test_points=None,
        test_description=None,
        test_use_in_statements=None,
        test_input_for_statements=None,
        test_output_for_statements=None,
        verify_input_output_for_statements=None,
        check_existing=None,
    ):
        return self._polygon.problem_save_test(
            self.id,
            testset,
            test_index,
            test_input,
            test_group,
            test_points,
            test_description,
            test_use_in_statements,
            test_input_for_statements,
            test_output_for_statements,
            verify_input_output_for_statements,
            check_existing,
        )

    def tests(self, testset):
        return self._polygon.problem_tests(self.id, testset)

    def save_test_group(
        self,
        testset,
        group,
        points_policy=None,
        feedback_policy=None,
        dependencies=None,
    ):
        return self._polygon.problem_save_test_group(
            self.id, testset, group, points_policy, feedback_policy, dependencies
        )

    def view_test_group(self, testset, group):
        return self._polygon.problem_view_test_group(testset, group)

    def view_file(self, type, name):
        return self._polygon.problem_view_file(self.id, type, name)

    def save_file(
        self, type, name, file, source_type=None, resource_advanced_properties=None
    ):
        return self._polygon.problem_save_file(
            self.id, type, name, file, source_type, resource_advanced_properties
        )

    def save_solution(self, name, file, source_type, tag, check_existing=None):
        return self._polygon.problem_save_solution(
            self.id, name, file, source_type, tag, check_existing
        )

    def checker(self):
        return self._polygon.problem_checker(self.id)

    def set_checker(self, checker):
        return self._polygon.problem_set_checker(self.id, checker)

    def validator(self):
        return self._polygon.problem_validator(self.id)

    def set_validator(self, validator):
        return self._polygon.problem_set_validator(self.id, validator)

    def interactor(self):
        return self._polygon.problem_interactor(self.id)

    def set_interactor(self, interactor):
        return self._polygon.problem_set_interactor(self.id, interactor)

    def solutions(self):
        return self._polygon.problem_solutions(self.id)

    def files(self):
        return self._polygon.problem_files(self.id)

    def files_resource(self):
        return self.files()[FileType.RESOURCE]

    def files_source(self):
        return self.files()[FileType.SOURCE]

    def files_aux(self):
        return self.files()[FileType.AUX]


class ProblemInfo:
    """
    Object representing Polygon problem info
    """

    # JSON field names
    _INPUT_FILE = 'inputFile'
    _OUTPUT_FILE = 'outputFile'
    _INTERACTIVE = 'interactive'
    _TIME_LIMIT = 'timeLimit'
    _MEMORY_LIMIT = 'memoryLimit'

    @classmethod
    def from_json(cls, problem_info_json):
        return cls(
            input_file=problem_info_json[ProblemInfo._INPUT_FILE],
            output_file=problem_info_json[ProblemInfo._OUTPUT_FILE],
            interactive=problem_info_json[ProblemInfo._INTERACTIVE],
            time_limit=problem_info_json[ProblemInfo._TIME_LIMIT],
            memory_limit=problem_info_json[ProblemInfo._MEMORY_LIMIT],
        )

    def __init__(
        self,
        input_file=None,
        output_file=None,
        interactive=None,
        time_limit=None,
        memory_limit=None,
    ):
        self.input_file = input_file
        self.output_file = output_file
        self.interactive = interactive
        self.time_limit = time_limit
        self.memory_limit = memory_limit


class Test:
    _INDEX = 'index'
    _MANUAL = 'manual'
    _INPUT = 'input'
    _DESCRIPTION = 'description'
    _USE_IN_STATEMENTS = 'useInStatements'
    _SCRIPT_LINE = 'scriptLine'
    _GROUP = 'group'
    _POINTS = 'points'
    _INPUT_FOR_STATEMENTS = 'inputForStatements'
    _OUTPUT_FOR_STATEMENTS = 'outputForStatements'
    _VERIFY_INPUT_OUTPUT_FOR_STATEMENTS = 'verifyInputOutputForStatements'

    @classmethod
    def from_json(cls, polygon, problem_id, testset, test_json):
        if test_json['manual']:
            return ManualTest.from_json(polygon, problem_id, testset, test_json)
        else:
            return GeneratedTest.from_json(polygon, problem_id, testset, test_json)

    def __init__(
        self,
        polygon,
        problem_id,
        testset,
        index,
        group,
        points,
        description,
        use_in_statements,
        input_for_statements,
        output_for_statements,
        verify_input_output_for_statements,
    ):
        self._polygon = polygon
        self._problem_id = problem_id
        self.testset = testset
        self.index = index
        self.group = group
        self.points = points
        self.description = description
        self.use_in_statements = use_in_statements
        self.input_for_statements = input_for_statements
        self.output_for_statements = output_for_statements
        self.verify_input_output_for_statements = verify_input_output_for_statements


class ManualTest(Test):
    @classmethod
    def from_json(cls, polygon, problem_id, testset, test_json):
        verify = test_json.get(Test._VERIFY_INPUT_OUTPUT_FOR_STATEMENTS, None)
        return cls(
            polygon=polygon,
            problem_id=problem_id,
            testset=testset,
            index=int(test_json[Test._INDEX]),
            input=test_json[Test._INPUT],
            group=test_json.get(Test._GROUP, ''),
            points=int(test_json.get(Test._POINTS, '0')),
            description=test_json.get(Test._DESCRIPTION, None),
            use_in_statements=test_json[Test._USE_IN_STATEMENTS],
            input_for_statements=test_json.get(Test._INPUT_FOR_STATEMENTS, None),
            output_for_statements=test_json.get(Test._OUTPUT_FOR_STATEMENTS, None),
            verify_input_output_for_statements=None
            if verify is None
            else (verify == 'true'),
        )

    def __init__(
        self,
        polygon,
        problem_id,
        testset,
        index,
        input,
        group=None,
        points=None,
        description=None,
        use_in_statements=None,
        input_for_statements=None,
        output_for_statements=None,
        verify_input_output_for_statements=None,
    ):
        super().__init__(
            polygon,
            problem_id,
            testset,
            index,
            group,
            points,
            description,
            use_in_statements,
            input_for_statements,
            output_for_statements,
            verify_input_output_for_statements,
        )
        self.input = input

    def save(self):
        return self._polygon.problem_save_test(
            self._problem_id,
            self.testset,
            self.index,
            self.input,
            self.group,
            self.points,
            self.description,
            self.use_in_statements,
            self.input_for_statements,
            self.output_for_statements,
            self.verify_input_output_for_statements,
        )


class GeneratedTest(Test):
    @classmethod
    def from_json(cls, polygon, problem_id, testset, test_json):
        verify = test_json.get(Test._VERIFY_INPUT_OUTPUT_FOR_STATEMENTS, None)
        return cls(
            polygon=polygon,
            problem_id=problem_id,
            testset=testset,
            index=int(test_json[Test._INDEX]),
            group=test_json.get(Test._GROUP, ''),
            points=int(test_json.get(Test._POINTS, '0')),
            description=test_json.get(Test._DESCRIPTION, None),
            use_in_statements=test_json[Test._USE_IN_STATEMENTS],
            script_line=test_json[Test._SCRIPT_LINE],
            input_for_statements=test_json.get(Test._INPUT_FOR_STATEMENTS, None),
            output_for_statements=test_json.get(Test._OUTPUT_FOR_STATEMENTS, None),
            verify_input_output_for_statements=None
            if verify is None
            else (verify == 'true'),
        )

    def __init__(
        self,
        polygon,
        problem_id,
        testset,
        index,
        group,
        points,
        description,
        use_in_statements,
        script_line,
        input_for_statements,
        output_for_statements,
        verify_input_output_for_statements,
    ):
        super().__init__(
            polygon,
            problem_id,
            testset,
            index,
            group,
            points,
            description,
            use_in_statements,
            input_for_statements,
            output_for_statements,
            verify_input_output_for_statements,
        )
        self.script_line = script_line


class TestGroup:
    """ """

    _NAME = 'name'
    _POINTS_POLICY = 'pointsPolicy'
    _FEEDBACK_POLICY = 'feedbackPolicy'
    _DEPENDENCIES = 'dependencies'

    @classmethod
    def from_json(cls, test_group_json):
        return cls(
            name=test_group_json[TestGroup._NAME],
            points_policy=PointsPolicy[test_group_json[TestGroup._POINTS_POLICY]],
            feedback_policy=FeedbackPolicy[test_group_json[TestGroup._FEEDBACK_POLICY]],
            dependencies=test_group_json[TestGroup._DEPENDENCIES],
        )

    def __init__(
        self, name, points_policy=None, feedback_policy=None, dependencies=None
    ):
        self.name = name
        if points_policy is not None and not isinstance(points_policy, PointsPolicy):
            raise TypeError(
                'Expected PointsPolicy instance for points_policy argument, but %s found'
                % type(points_policy)
            )
        if feedback_policy is not None and not isinstance(
            feedback_policy, FeedbackPolicy
        ):
            raise TypeError(
                'Expected FeedbackPolicy instance for feedback_policy argument, but %s found'
                % type(feedback_policy)
            )
        self.points_policy = points_policy
        self.feedback_policy = feedback_policy
        self.dependencies = dependencies


class Statement:
    """
    Object: representing Polygon problem statement
    """

    _ENCODING = 'encoding'
    _NAME = 'name'
    _LEGEND = 'legend'
    _INPUT = 'input'
    _OUTPUT = 'output'
    _SCORING = 'scoring'
    _INTERACTION = 'interaction'
    _NOTES = 'notes'
    _TUTORIAL = 'tutorial'

    @classmethod
    def from_json(cls, statement_json):
        return cls(
            encoding=statement_json[Statement._ENCODING],
            name=statement_json[Statement._NAME],
            legend=statement_json[Statement._LEGEND],
            input=statement_json[Statement._INPUT],
            output=statement_json[Statement._OUTPUT],
            scoring=statement_json.get(Statement._SCORING, None),
            interaction=statement_json.get(Statement._INTERACTION, None),
            notes=statement_json[Statement._NOTES],
            tutorial=statement_json[Statement._TUTORIAL],
        )

    def __init__(
        self,
        encoding=None,
        name=None,
        legend=None,
        input=None,
        output=None,
        scoring=None,
        interaction=None,
        notes=None,
        tutorial=None,
    ):
        self.encoding = encoding
        self.name = name
        self.legend = legend
        self.input = input
        self.output = output
        self.scoring = scoring
        self.interaction = interaction
        self.notes = notes
        self.tutorial = tutorial


class Solution:
    """
    Object: representing Polygon problem solution
    """

    _NAME = 'name'
    _MODIFICATION_TIME_SECONDS = 'modificationTimeSeconds'
    _LENGTH = 'length'
    _SOURCE_TYPE = 'sourceType'
    _TAG = 'tag'

    @classmethod
    def from_json(cls, solution_json):
        return cls(
            name=solution_json[Solution._NAME],
            modification_time_seconds=solution_json[
                Solution._MODIFICATION_TIME_SECONDS
            ],
            length=solution_json[Solution._LENGTH],
            source_type=solution_json[Solution._SOURCE_TYPE],
            tag=SolutionTag[solution_json[Solution._TAG]],
        )

    def __init__(self, name, modification_time_seconds, length, source_type, tag):
        self.name = name
        self.modification_time_seconds = modification_time_seconds
        self.length = length
        self.source_type = source_type
        self.tag = tag


class File:
    """
    Object: representing Polygon file (resource, source, or aux, not solutions)
    """

    _NAME = 'name'
    _MODIFICATION_TIME_SECONDS = 'modificationTimeSeconds'
    _LENGTH = 'length'
    _SOURCE_TYPE = 'sourceType'
    _RESOURCE_ADVANCED_PROPERTIES = 'resourceAdvancedProperties'

    @classmethod
    def from_json(cls, file_json):
        return cls(
            name=file_json[File._NAME],
            modification_time_seconds=file_json[File._MODIFICATION_TIME_SECONDS],
            length=file_json[File._LENGTH],
            source_type=file_json.get(File._SOURCE_TYPE, None),
            resource_advanced_properties=ResourceAdvancedProperties.from_json(
                file_json[File._RESOURCE_ADVANCED_PROPERTIES]
            )
            if File._RESOURCE_ADVANCED_PROPERTIES in file_json.keys()
            else None,
        )

    def __init__(
        self,
        name,
        modification_time_seconds,
        length,
        source_type,
        resource_advanced_properties,
    ):
        self.name = name
        self.modification_time_seconds = modification_time_seconds
        self.length = length
        self.source_type = source_type
        self.resource_advanced_properties = resource_advanced_properties


class ResourceAdvancedProperties:
    """ """

    _FOR_TYPES = 'forTypes'
    _MAIN = 'main'
    _STAGES = 'stages'
    _ASSETS = 'assets'

    @classmethod
    def from_json(cls, resource_advanced_properties):
        return cls(
            for_types=resource_advanced_properties[
                ResourceAdvancedProperties._FOR_TYPES
            ],
            main=resource_advanced_properties[ResourceAdvancedProperties._MAIN],
            stages=[
                Stage[stage]
                for stage in resource_advanced_properties[
                    ResourceAdvancedProperties._STAGES
                ]
            ],
            assets=[
                Asset[asset]
                for asset in resource_advanced_properties[
                    ResourceAdvancedProperties._ASSETS
                ]
            ],
        )

    def __init__(self, for_types=None, main=None, stages=None, assets=None):
        self.for_types = for_types
        self.main = main
        self.stages = stages
        self.assets = assets


ResourceAdvancedProperties.DELETE = ResourceAdvancedProperties(for_types='')


class Request:
    """
    Request to Polygon API.
    TODO: update
    Usage example:
    >>> request = Request(Request.CONTEST_PROBLEMS,
    ...                   args={'contestId': contest_id})
    >>> response = request.issue()
    >>> problems = response["result"]
    """

    @staticmethod
    def _encoded_args(args):
        return [
            (Request._value_to_utf8_bytes(key), Request._value_to_utf8_bytes(value))
            for key, value in args
        ]

    @staticmethod
    def _value_to_utf8_bytes(value):
        if isinstance(value, bytes):
            return value
        if isinstance(value, bytearray):
            return bytes(value)
        return str(value).encode('utf-8')

    def __init__(self, config, method_name, args=None):
        if args is None:
            args = {}
        args = {key: value for key, value in args.items() if value is not None}
        for key in args:
            if isinstance(args[key], bool):
                args[key] = 'true' if args[key] else 'false'
        self.args = []
        for key, value in args.items():
            if isinstance(value, list):
                for it in value:
                    self.args.append((key, it))
            else:
                self.args.append((key, value))
        self.config = config
        self.method_name = method_name

    def issue(self):
        """Issues request and returns parsed JSON response"""

        try:
            return Response(json.loads(self.issue_raw()))
        except json.JSONDecodeError:
            console.console.print(
                '[error]Failed to parse JSON response from Polygon API. Please check if Polygon is down.[/error]',
            )
            raise typer.Exit(1)

    def issue_raw(self):
        """Issues request and returns raw response (useful e.g. for files)"""

        args = list(self.args)
        args.append(('apiKey', self.config.api_key))
        args.append(('time', str(int(time.time()))))
        args = Request._encoded_args(args)
        args.append(
            (
                b'apiSig',
                self.get_api_signature(
                    args, Request._value_to_utf8_bytes(self.config.api_secret)
                ),
            )
        )
        response = requests.post(self.config.api_url + self.method_name, files=args)
        return response.text

    def get_api_signature(self, args, api_secret):
        rand_bit = Request._value_to_utf8_bytes(
            ''.join(
                random.choice(string.ascii_lowercase + string.digits) for _ in range(6)
            )
        )

        arg_tuples = list(sorted(args))
        args_bit = b'&'.join(key + b'=' + value for key, value in arg_tuples)
        api_signature_string = (
            rand_bit
            + b'/'
            + self.method_name.encode('utf-8')
            + b'?'
            + args_bit
            + b'#'
            + api_secret
        )
        api_signature = rand_bit + Request._value_to_utf8_bytes(
            hashlib.sha512(api_signature_string).hexdigest()
        )
        return api_signature


class Response:
    """ """

    # JSON field names
    FIELD_STATUS = 'status'
    FIELD_COMMENT = 'comment'
    FIELD_RESULT = 'result'

    # Status values
    STATUS_OK = 'OK'
    STATUS_FAILED = 'FAILED'
    STATUS_UNKNOWN = 'UNKNOWN'

    def __init__(self, response_json):
        self.status = response_json.get(Response.FIELD_STATUS)
        if self.status not in [Response.STATUS_OK, Response.STATUS_FAILED]:
            self.status = Response.STATUS_UNKNOWN

        self.comment = response_json.get(Response.FIELD_COMMENT, '')
        self.result = response_json.get(Response.FIELD_RESULT)


class RequestConfig:
    def __init__(self, api_url, api_key, api_secret):
        if not api_url.endswith('/'):
            api_url += '/'
        self.api_url = api_url
        self.api_key = api_key
        self.api_secret = api_secret


class PolygonRequestFailedException(Exception):
    """Exception raised when Polygon returns FAILED as request status"""

    def __init__(self, comment):
        self.comment = comment


class PointsPolicy(Enum):
    COMPLETE_GROUP = 0
    EACH_TEST = 1

    def __str__(self):
        return self.name


class FeedbackPolicy(Enum):
    NONE = 0
    POINTS = 1
    ICPC = 2
    COMPLETE = 3

    def __str__(self):
        return self.name


class FileType(Enum):
    RESOURCE = 0
    SOURCE = 1
    AUX = 2

    def __str__(self):
        return self.name.lower()


class SolutionTag(Enum):
    MA = 0  # Main correct solution
    OK = 1  # Accepted
    RJ = 2  # Rejected, Incorrect
    TL = 3  # Time limit exceeded
    TO = 4  # Time limit exceeded or accepted
    WA = 5  # Wrong answer
    PE = 6  # Presentation error
    ML = 7  # Memory limit exceeded
    RE = 8  # Runtime error
    TM = 9  # Time limit exceeded or memory limit exceeded
    NR = 10  # Do not run

    def __str__(self):
        return self.name


class Asset(Enum):
    VALIDATOR = 0
    INTERACTOR = 1
    CHECKER = 2
    SOLUTION = 3

    def __str__(self):
        return self.name


class Stage(Enum):
    COMPILE = 0
    RUN = 1

    def __str__(self):
        return self.name



import contextlib
import functools
import pathlib
from typing import List, Optional

import typer

from rbx import console, utils
from rbx.box.sanitizers import warning_stack
from rbx.utils import new_cd


def find_package(
    root: pathlib.Path = pathlib.Path(), consider_presets: bool = False
) -> Optional[pathlib.Path]:
    root = utils.abspath(root)

    def has_file():
        problem_yaml_path = root / 'problem.rbx.yml'
        contest_yaml_path = root / 'contest.rbx.yml'
        preset_yaml_path = root / 'preset.rbx.yml'
        return (
            problem_yaml_path.is_file()
            or contest_yaml_path.is_file()
            or (consider_presets and preset_yaml_path.is_file())
        )

    while root != pathlib.PosixPath('/') and not has_file():
        root = root.parent
    if not has_file():
        return None
    return root


def find_all_ancestor_packages(
    root: pathlib.Path = pathlib.Path(),
) -> List[pathlib.Path]:
    packages = []
    while (pkg := find_package(root)) is not None:
        packages.append(pkg)
        root = pkg.parent
    return packages


def is_problem_package(root: pathlib.Path = pathlib.Path()) -> bool:
    dir = find_package(root)
    if dir is None:
        return False
    return (dir / 'problem.rbx.yml').is_file()


def is_contest_package(root: pathlib.Path = pathlib.Path()) -> bool:
    dir = find_package(root)
    if dir is None:
        return False
    return (dir / 'contest.rbx.yml').is_file()


def is_preset_package(root: pathlib.Path = pathlib.Path()) -> bool:
    dir = find_package(root, consider_presets=True)
    if dir is None:
        return False
    return (dir / 'preset.rbx.yml').is_file()


def within_closest_package(func, consider_presets: bool = False):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        package = find_package(consider_presets=consider_presets)
        if package is None:
            console.console.print('[error]No rbx package found.[/error]')
            raise typer.Exit(1)
        # Get deepest package.
        with new_package_cd(package):
            return func(*args, **kwargs)

    return wrapper


def within_closest_wrapper(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        package = find_package(consider_presets=True)
        if package is None:
            console.console.print('[error]No rbx package found.[/error]')
            raise typer.Exit(1)
        # Get deepest package.
        with new_package_cd(package):
            return func(*args, **kwargs)

    return wrapper


@contextlib.contextmanager
def new_package_cd(x: pathlib.Path):
    with new_cd(x):
        yield
        warning_stack.print_warning_stack_report()



import dataclasses
import pathlib
import re
import resource
import shlex
import sys
from enum import Enum
from pathlib import PosixPath
from typing import Any, List, Optional, Tuple, Union

import rich
import rich.text
import typer
from pydantic import BaseModel

from rbx import console
from rbx.box import download, global_package, package, setter_config, state
from rbx.box.environment import (
    BaseCompilationConfig,
    CompilationConfig,
    EnvironmentLanguage,
    ExecutionConfig,
    FileMapping,
    get_compilation_config,
    get_execution_config,
    get_file_mapping,
    get_language,
    get_language_by_extension,
    get_mapped_command,
    get_mapped_commands,
    get_sandbox_params_from_config,
    merge_execution_configs,
)
from rbx.box.formatting import get_formatted_memory
from rbx.box.remote import is_path_remote
from rbx.box.sanitizers import warning_stack
from rbx.box.schema import CodeItem, Solution
from rbx.grading import grading_context, profiling, steps, steps_with_caching
from rbx.grading.judge.sandbox import SandboxParams
from rbx.grading.steps import (
    DigestHolder,
    DigestOrDest,
    DigestOrSource,
    GradingArtifacts,
    GradingFileInput,
    GradingFileOutput,
    RunLog,
    RunLogMetadata,
    get_exe_from_command,
    is_cpp_command,
    is_cxx_command,
    maybe_get_bits_stdcpp_for_commands,
)
from rbx.utils import is_arm

MERGED_CAPTURE_FILENAME = 'merged_capture.pio'

PASSTHROUGH_VARIABLES = ['initialMemory', 'memory']


class SanitizationLevel(Enum):
    NONE = 0
    PREFER = 1
    FORCE = 2

    def should_sanitize(self) -> bool:
        cfg = setter_config.get_setter_config()
        if cfg.sanitizers.enabled or state.STATE.sanitized:
            return self.value >= SanitizationLevel.PREFER.value
        return self.value >= SanitizationLevel.FORCE.value


class CompilationMetadata(BaseModel):
    is_sanitized: bool


def substitute_commands(commands: List[str], sanitized: bool = False) -> List[str]:
    cfg = setter_config.get_setter_config()
    return [cfg.substitute_command(command, sanitized) for command in commands]


def get_extension(code: CodeItem) -> str:
    path: pathlib.Path = PosixPath(code.path)
    return path.suffix[1:]


def find_language(code: CodeItem) -> EnvironmentLanguage:
    if code.language is not None:
        return get_language(code.language)
    return get_language_by_extension(get_extension(code))


def find_language_name(code: CodeItem) -> str:
    return find_language(code).name


def is_executable_sanitized(executable: DigestOrSource) -> bool:
    if executable.digest is None:
        return False
    if executable.digest.value is None:
        return False
    cacher = package.get_file_cacher()
    desc = cacher.get_metadata(
        executable.digest.value, 'compilation', CompilationMetadata
    )
    if desc is None:
        return False
    return desc.is_sanitized


def add_sanitizer_flags_to_command(command: str) -> str:
    if is_cxx_command(command):
        return command + ' -fsanitize=address,undefined -fno-omit-frame-pointer -g'
    return command


def add_sanitizer_flags(commands: List[str]) -> List[str]:
    return [add_sanitizer_flags_to_command(command) for command in commands]


CXX_WARNING_FLAGS = (
    '-Wall -Wshadow -Wno-unused-result -Wno-sign-compare -Wno-char-subscripts'
)


def add_color_flags_to_command(command: str) -> str:
    if is_cxx_command(command):
        return command + ' -fdiagnostics-color=always'
    return command


def add_warning_flags_to_command(command: str) -> str:
    if is_cxx_command(command):
        return command + ' ' + CXX_WARNING_FLAGS
    return command


def add_arm_flags_to_command(command: str) -> str:
    # Disable FMA in ARM.
    if is_cxx_command(command) and is_arm():
        return command + ' -ffp-contract=off'
    return command


def add_cpp_flags(commands: List[str], force_warnings: bool) -> List[str]:
    cfg = setter_config.get_setter_config()
    commands = [add_color_flags_to_command(command) for command in commands]
    commands = [add_arm_flags_to_command(command) for command in commands]
    if cfg.warnings.enabled or force_warnings:
        return [add_warning_flags_to_command(command) for command in commands]
    return commands


def _add_warning_pragmas(code: str) -> str:
    flags = CXX_WARNING_FLAGS.split()
    pragma_lines = '\n'.join(
        [
            f'#pragma GCC diagnostic ignored "{flag}"'
            for flag in flags
            if not flag.startswith('-Wno-')
        ]
    )

    return re.sub(
        r'^(#include[^\n]*)',
        '#pragma GCC diagnostic push\n'
        + pragma_lines
        + '\n\\1'
        + '\n#pragma GCC diagnostic pop\n',
        code,
        flags=re.MULTILINE,
    )


def _add_warning_pragmas_around(code: str) -> str:
    flags = CXX_WARNING_FLAGS.split()
    pragma_lines = '\n'.join(
        [
            f'#pragma GCC diagnostic ignored "{flag}"'
            for flag in flags
            if not flag.startswith('-Wno-')
        ]
    )

    return (
        '#pragma GCC diagnostic push\n'
        + pragma_lines
        + '\n'
        + code
        + '\n'
        + '#pragma GCC diagnostic pop\n'
    )


def _ignore_warning_in_cxx_input(input: GradingFileInput):
    if input.src is None or input.src.suffix not in ('.h', '.hpp'):
        return
    preprocessed_path = package.get_problem_preprocessed_path(input.src)
    preprocessed_path.write_text(_add_warning_pragmas_around(input.src.read_text()))
    input.src = preprocessed_path


def _get_java_class_name(compilable_path: pathlib.Path) -> Optional[str]:
    if compilable_path.suffix != '.java':
        return None

    java_content = compilable_path.read_text()
    regex = re.compile(r'public\s+class\s+([A-Za-z0-9_$]+)')
    match = regex.search(java_content)
    if match is None:
        return compilable_path.name
    return match.group(1)


def _get_code_variables(code: CodeItem, language: str) -> dict[str, Any]:
    res = {'source': code.path.name, 'language': language}
    java_klass = _get_java_class_name(code.path)
    if java_klass is not None:
        res['javaClass'] = java_klass
    return res


def maybe_rename_java_class(
    compilable_path: pathlib.Path,
    file_mapping: FileMapping,
) -> pathlib.Path:
    if file_mapping.compilable != 'Main.java':
        # This function is only kept here for backwards compatibility.
        # It should only be run when the file mapping is set to 'Main.java'.
        # TODO: remove this function.
        return compilable_path

    mapped_path = PosixPath(file_mapping.compilable)
    if mapped_path.suffix != '.java':
        return compilable_path
    import re

    cls_name = mapped_path.stem

    java_content = compilable_path.read_text()
    regex = re.compile(r'public\s+class\s+[A-Za-z0-9_$]+([^A-Za-z0-9_$])')
    match = regex.search(java_content)
    if match is None:
        console.console.print(
            f'[error]Java public class not found in file: [item]{compilable_path}[/item][/error]'
        )
        raise typer.Exit(1)

    new_content = regex.sub(f'public class {cls_name}\\1', java_content)
    if new_content == java_content:
        return compilable_path

    preprocessed_path = package.get_problem_preprocessed_path(compilable_path)
    preprocessed_path.write_text(new_content)
    return preprocessed_path


def _format_stack_limit(limit: int) -> str:
    if limit == resource.RLIM_INFINITY:
        return 'unlimited'
    return get_formatted_memory(limit)


def _check_stack_limit():
    cfg = setter_config.get_setter_config()
    if not cfg.judging.check_stack:
        return
    if not state.STATE.run_through_cli:
        return
    if sys.platform != 'darwin':
        return
    soft, hard = resource.RLIM_INFINITY, resource.RLIM_INFINITY

    TARGET = 256 * 1024 * 1024  # 256 MiB
    try:
        soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
    except Exception:
        pass

    if soft != hard and soft != resource.RLIM_INFINITY and soft < TARGET:
        soft_fmt = _format_stack_limit(soft)
        hard_fmt = _format_stack_limit(hard)
        console.console.print(
            f'[error]Stack limit is too low (limit is set as [item]{soft_fmt}[/item], but configured user capacity is [item]{hard_fmt}[/item]).[/error]'
        )
        console.console.print(
            '[error]It is not safe to develop problems in [item]rbx[/item] with this configuration.[/error]'
        )
        console.console.print(
            'To solve this, add the following lines to the end of your [item]~/.bashrc[/item] or [item]~/.zshrc[/item] file (or equivalent shell configuration file):'
        )

        target_text = TARGET
        if hard != resource.RLIM_INFINITY:
            target_text = min(hard, TARGET)
        console.console.print(
            f"""
```
function rbx() {{
    local rbx_bin=`bash -c "type -P rbx"`
    ulimit -s {target_text // 1024} && $rbx_bin "$@"
}}
```
        """
        )
        console.console.print()
        console.console.print(
            'You can read more about this in [item]https://rsalesc.github.io/rbx/stack-limit/[/item].'
        )
        raise typer.Exit(1)


@dataclasses.dataclass
class PreparedRun:
    command: str
    sandbox_params: SandboxParams
    artifacts: GradingArtifacts
    sanitized: bool

    file_mapping: FileMapping
    metadata: RunLogMetadata


def _prepare_run(
    code: CodeItem,
    executable: DigestOrSource,
    stdin: Optional[DigestOrSource] = None,
    stdout: Optional[DigestOrDest] = None,
    stderr: Optional[DigestOrDest] = None,
    inputs: Optional[List[GradingFileInput]] = None,
    outputs: Optional[List[GradingFileOutput]] = None,
    extra_args: Optional[str] = None,
    extra_config: Optional[ExecutionConfig] = None,
    retry_index: Optional[int] = None,
    file_prefix: Optional[str] = None,
):
    language = find_language_name(code)
    execution_options = get_execution_config(
        language, solution=isinstance(code, Solution)
    )
    if extra_config is not None:
        execution_options = merge_execution_configs([execution_options, extra_config])
    code_variables = _get_code_variables(code, language)
    file_mapping = get_file_mapping(language, code_variables, file_prefix)
    sandbox_params = get_sandbox_params_from_config(execution_options.sandbox)

    # Sanitization parameters.
    sanitized = False
    if is_executable_sanitized(executable):
        # Remove any memory constraints for a sanitized executable.
        # Sanitizers are known to be memory-hungry.
        sandbox_params.address_space = None

        # Reset timeout configs since sanitizers are known to be time-hungry.
        sandbox_params.timeout = None
        sandbox_params.wallclock_timeout = None
        sanitized = True

    sandbox_params.set_stdall(
        stdin=PosixPath(file_mapping.input) if stdin is not None else None,
        stdout=PosixPath(file_mapping.output) if stdout is not None else None,
        stderr=PosixPath(file_mapping.error)
        if stderr is not None or sanitized
        else None,
    )

    assert execution_options.command
    command = get_mapped_command(
        execution_options.command,
        file_mapping,
        code_variables,
        passthrough=PASSTHROUGH_VARIABLES,
    )
    command = substitute_commands([command], sanitized=sanitized)[0]

    if extra_args is not None:
        splitted_command = shlex.split(command)
        splitted_command.extend(shlex.split(extra_args))
        command = shlex.join(splitted_command)

    artifacts = GradingArtifacts()
    artifacts.inputs.append(
        GradingFileInput(
            **executable.expand(),
            dest=PosixPath(file_mapping.executable),
            executable=True,
        )
    )
    if stdin is not None:
        artifacts.inputs.append(
            GradingFileInput(
                **stdin.expand(),
                dest=PosixPath(file_mapping.input),
            )
        )
    if stdout is not None:
        artifacts.outputs.append(
            GradingFileOutput(
                src=PosixPath(file_mapping.output),
                **stdout.expand(),
                touch=True,
            )
        )
    if stderr is not None:
        artifacts.outputs.append(
            GradingFileOutput(
                src=PosixPath(file_mapping.error),
                **stderr.expand(),
                touch=True,
            )
        )
    if inputs:
        artifacts.inputs.extend(inputs)
    if outputs:
        artifacts.outputs.extend(outputs)

    return PreparedRun(
        command=command,
        sandbox_params=sandbox_params,
        artifacts=artifacts,
        sanitized=sanitized,
        file_mapping=file_mapping,
        metadata=RunLogMetadata(
            language=language,
            is_sanitized=sanitized,
            timeLimit=sandbox_params.timeout,
            memoryLimit=sandbox_params.address_space,
            limits=execution_options.problemLimits,
            retryIndex=retry_index,
        ),
    )


def _should_precompile(commands: List[str]) -> bool:
    return any(is_cpp_command(command) for command in commands)


def _precompile_header(
    compilation_options: Union[CompilationConfig, BaseCompilationConfig],
    sanitized: SanitizationLevel,
    sandbox_params: SandboxParams,
    artifacts: GradingArtifacts,
    input_artifact: GradingFileInput,
    force_warnings: bool = False,
    verbose: bool = False,
    include_other_headers: bool = False,
) -> GradingFileInput:
    """
    Precompile a header file (.h).

    Assumes input artifact is a header file (.h) and compilation commands are C++.
    """
    assert compilation_options.commands is not None

    sandbox = global_package.get_global_sandbox()
    dependency_cache = global_package.get_global_dependency_cache()

    # TODO: deduplicate code with compile_item.
    commands = get_mapped_commands(
        compilation_options.commands,
        FileMapping(
            compilable='precompilable.h',
            executable='precompilable.h.gch',
        ),
        passthrough=PASSTHROUGH_VARIABLES,
    )
    commands = add_cpp_flags(commands, force_warnings)
    commands = substitute_commands(commands, sanitized=sanitized.should_sanitize())

    if sanitized.should_sanitize():
        commands = add_sanitizer_flags(commands)

    precompilation_artifacts = GradingArtifacts()

    # Keep only header files.
    if include_other_headers:
        precompilation_artifacts.inputs = [
            input
            for input in artifacts.inputs
            if input.src is not None and input.src.suffix == '.h'
        ]
    precompilation_artifacts.inputs.append(
        GradingFileInput(
            src=input_artifact.src,
            dest=PosixPath('precompilable.h'),
        )
    )

    # Pull only the precompiled header file.
    precompiled_digest = DigestHolder()
    precompilation_artifacts.outputs.append(
        GradingFileOutput(
            src=PosixPath('precompilable.h.gch'),
            digest=precompiled_digest,
        )
    )

    with profiling.PushContext('code.precompile_header'):
        try:
            steps_with_caching.compile(
                commands,
                params=sandbox_params,
                artifacts=precompilation_artifacts,
                sandbox=sandbox,
                dependency_cache=dependency_cache,
            )
        except steps.CompilationError as e:
            e.print(
                f'[error]Failed to precompile header file: [item]{input_artifact.src}[/item][/error]'
            )
            raise

        if verbose:
            console.console.print(
                f'[status]Precompiled header file: [item]{input_artifact.src}[/item]'
            )

            if (
                precompilation_artifacts.logs is not None
                and precompilation_artifacts.logs.preprocess is not None
            ):
                for log in precompilation_artifacts.logs.preprocess:
                    console.console.print(
                        f'[status]Command:[/status] {log.get_command()}'
                    )
                    console.console.print(
                        f'[status]Summary:[/status] {log.get_summary()}'
                    )

    assert precompiled_digest.value is not None

    digest_path = dependency_cache.cacher.path_for_symlink(precompiled_digest.value)
    if digest_path is not None and digest_path.is_file():
        # If storage backend supports symlinks, use it as the grading input.
        input = DigestOrSource.create(digest_path)
    else:
        # Otherwise, copy the file to the local cache, transiently.
        local_cacher = package.get_file_cacher()
        with dependency_cache.cacher.get_file(precompiled_digest.value) as f:
            with grading_context.cache_level(
                grading_context.CacheLevel.CACHE_TRANSIENTLY
            ):
                input = DigestOrSource.create(local_cacher.put_file_from_fobj(f))

    res = GradingFileInput(
        **input.expand(),
        dest=input_artifact.dest.with_suffix('.h.gch'),
        # Do not track fingerprint of the precompiled header file,
        # trust the compilation step above.
        hash=False,
    )
    return res


# Compile code item and return its digest in the storage.
def compile_item(
    code: CodeItem,
    sanitized: SanitizationLevel = SanitizationLevel.PREFER,
    force_warnings: bool = False,
    verbose: bool = False,
    precompile: bool = True,
) -> str:
    _check_stack_limit()

    compilable_path = PosixPath(code.path)

    if not compilable_path.is_file():
        console.console.print(
            f'[error]Compilation file not found: [item]{compilable_path}[/item][/error]'
        )
        raise typer.Exit(1)

    language = find_language_name(code)
    compilation_options = get_compilation_config(
        language, solution=isinstance(code, Solution)
    )
    code_variables = _get_code_variables(code, language)
    file_mapping = get_file_mapping(language, code_variables)
    dependency_cache = package.get_dependency_cache()
    sandbox = package.get_singleton_sandbox()
    sandbox_params = get_sandbox_params_from_config(compilation_options.sandbox)
    sandbox_params.set_env['CLICOLOR_FORCE'] = '1'

    if not compilation_options.commands:
        # Language is not compiled.
        return sandbox.file_cacher.put_file_from_path(compilable_path)

    commands = get_mapped_commands(
        compilation_options.commands,
        file_mapping,
        code_variables,
        passthrough=PASSTHROUGH_VARIABLES,
    )
    commands = add_cpp_flags(commands, force_warnings)
    commands = substitute_commands(commands, sanitized=sanitized.should_sanitize())

    if sanitized.should_sanitize():
        commands = add_sanitizer_flags(commands)

        # Remove any memory constraints for a sanitized executable.
        # Sanitizers are known to be memory-hungry.
        sandbox_params.address_space = None

        # Reset timeout configs since sanitizers are known to be time-hungry.
        sandbox_params.timeout = None
        sandbox_params.wallclock_timeout = None

    compiled_digest = DigestHolder()

    artifacts = GradingArtifacts()
    artifacts.inputs.extend(
        GradingFileInput(src=src, dest=dest)
        for src, dest in package.get_compilation_files(code)
    )

    download.maybe_add_testlib(code, artifacts)
    download.maybe_add_jngen(code, artifacts)
    download.maybe_add_rbx_header(code, artifacts)
    compilable_path = maybe_rename_java_class(compilable_path, file_mapping)
    artifacts.inputs.append(
        GradingFileInput(src=compilable_path, dest=PosixPath(file_mapping.compilable))
    )

    artifacts.outputs.append(
        GradingFileOutput(
            src=PosixPath(file_mapping.executable)
            if not compilation_options.passthrough
            else PosixPath(file_mapping.compilable),
            digest=compiled_digest,
            executable=True,
        )
    )

    for input in artifacts.inputs:
        _ignore_warning_in_cxx_input(input)

    # Add system bits/stdc++.h to the compilation.
    bits_artifact = maybe_get_bits_stdcpp_for_commands(commands)
    if bits_artifact is not None:
        artifacts.inputs.append(bits_artifact)
        commands = [
            command + ' -I.'
            for command in commands
            if is_cxx_command(get_exe_from_command(command))
        ]

    # Precompile C++ interesting header files.
    if precompile and _should_precompile(commands):
        with profiling.Profiler('code.precompile'):
            precompilation_inputs = []
            for input in artifacts.inputs:
                if (
                    input.src is not None
                    and input.src.suffix == '.h'
                    and input.dest.name in ['stdc++.h', 'jngen.h', 'testlib.h']
                ):
                    precompilation_inputs.append(
                        _precompile_header(
                            compilation_options,
                            sanitized,
                            sandbox_params,
                            artifacts,
                            input,
                            force_warnings,
                            verbose=False,
                        )
                    )
            if precompilation_inputs:
                artifacts.inputs.extend(precompilation_inputs)

    with profiling.Profiler('code.compile'):
        # Compile the code.
        # Do not cache remote solutions.
        with grading_context.cache_level(
            grading_context.CacheLevel.NO_CACHE,
            when=lambda: is_path_remote(code.path),
        ):
            try:
                steps_with_caching.compile(
                    commands,
                    params=sandbox_params,
                    artifacts=artifacts,
                    sandbox=sandbox,
                    dependency_cache=dependency_cache,
                )
            except steps.CompilationError as e:
                e.print(f'[error]Failed to compile item: {code.href()}[/error]')
                raise

    assert compiled_digest.value is not None

    if verbose and artifacts.logs is not None and artifacts.logs.preprocess is not None:
        console.console.print(f'[status]Compiled item: {code.href()}')
        for log in artifacts.logs.preprocess:
            console.console.print(f'[status]Command:[/status] {log.get_command()}')
            console.console.print(f'[status]Summary:[/status] {log.get_summary()}')
            console.console.print(rich.text.Text.from_ansi(log.log), style='default')

    # Write compiler warnings.
    cfg = setter_config.get_setter_config()
    if (
        (cfg.warnings.enabled or force_warnings)
        and artifacts.logs is not None
        and artifacts.logs.preprocess is not None
    ):
        any_warning = any(log.warnings for log in artifacts.logs.preprocess)
        if any_warning:
            warning_stack.get_warning_stack().add_warning(code)

    # Create sentinel to indicate this executable is sanitized.
    cacher = package.get_file_cacher()
    with grading_context.cache_level(
        grading_context.CacheLevel.NO_CACHE,
        when=lambda: is_path_remote(code.path),
    ):
        if sanitized.should_sanitize():
            cacher.set_metadata(
                compiled_digest.value,
                'compilation',
                CompilationMetadata(is_sanitized=True),
            )
        else:
            cacher.set_metadata(compiled_digest.value, 'compilation', None)

    return compiled_digest.value


async def run_item(
    code: CodeItem,
    executable: DigestOrSource,
    stdin: Optional[DigestOrSource] = None,
    stdout: Optional[DigestOrDest] = None,
    stderr: Optional[DigestOrDest] = None,
    inputs: Optional[List[GradingFileInput]] = None,
    outputs: Optional[List[GradingFileOutput]] = None,
    extra_args: Optional[str] = None,
    extra_config: Optional[ExecutionConfig] = None,
    retry_index: Optional[int] = None,
) -> Optional[RunLog]:
    _check_stack_limit()

    dependency_cache = package.get_dependency_cache()

    prepared = _prepare_run(
        code,
        executable,
        stdin,
        stdout,
        stderr,
        inputs,
        outputs,
        extra_args,
        extra_config,
        retry_index,
    )

    with profiling.PushContext('code.run_item'):
        # Do not cache remote solutions.
        with grading_context.cache_level(
            grading_context.CacheLevel.NO_CACHE,
            when=lambda: is_path_remote(code.path),
        ):
            run_log = await steps_with_caching.run(
                prepared.command,
                params=prepared.sandbox_params,
                sandbox=package.get_singleton_sandbox(),
                artifacts=prepared.artifacts,
                dependency_cache=dependency_cache,
                metadata=prepared.metadata,
            )

    # Find sanitizer logs.
    if run_log is not None and run_log.warnings:
        assert prepared.sandbox_params.stderr_file is not None
        stderr_output = prepared.artifacts.get_output_file_for_src(
            prepared.sandbox_params.stderr_file
        )
        if stderr_output is not None:
            warning_stack.get_warning_stack().add_sanitizer_warning(
                package.get_file_cacher(), code, stderr_output
            )
    return run_log


@dataclasses.dataclass
class CommunicationItem:
    code: CodeItem
    executable: DigestOrSource
    file_prefix: str
    stderr: Optional[DigestOrDest] = None
    inputs: Optional[List[GradingFileInput]] = None
    outputs: Optional[List[GradingFileOutput]] = None
    extra_args: Optional[str] = None
    extra_config: Optional[ExecutionConfig] = None
    capture: Optional[DigestOrDest] = None

    def prepare(self) -> PreparedRun:
        return _prepare_run(
            self.code,
            self.executable,
            stdout=self.capture,
            stderr=self.stderr,
            inputs=self.inputs,
            outputs=self.outputs,
            extra_args=self.extra_args,
            extra_config=self.extra_config,
            file_prefix=self.file_prefix,
        )


async def run_communication(
    interactor: CommunicationItem,
    solution: CommunicationItem,
    merged_capture: Optional[DigestOrDest] = None,
    line_capture: bool = False,
    retry_index: Optional[int] = None,
) -> Tuple[Optional[RunLog], Optional[RunLog]]:
    interactor_prepared = interactor.prepare()
    solution_prepared = solution.prepare()

    # Prepare retry index.
    interactor_prepared.metadata.retryIndex = retry_index
    solution_prepared.metadata.retryIndex = retry_index

    grading_artifacts = GradingArtifacts()
    grading_artifacts.inputs.extend(interactor_prepared.artifacts.inputs)
    grading_artifacts.outputs.extend(interactor_prepared.artifacts.outputs)
    grading_artifacts.inputs.extend(solution_prepared.artifacts.inputs)
    grading_artifacts.outputs.extend(solution_prepared.artifacts.outputs)

    merged_capture_path: Optional[pathlib.Path] = None
    if merged_capture is not None:
        merged_capture_path = pathlib.Path(MERGED_CAPTURE_FILENAME)
        grading_artifacts.outputs.append(
            GradingFileOutput(
                src=merged_capture_path,
                **merged_capture.expand(),
            )
        )

    interactor_run_params = steps.CoordinatedRunParams(
        command=interactor_prepared.command,
        params=interactor_prepared.sandbox_params,
        metadata=interactor_prepared.metadata,
    )
    solution_run_params = steps.CoordinatedRunParams(
        command=solution_prepared.command,
        params=solution_prepared.sandbox_params,
        metadata=solution_prepared.metadata,
    )

    # Do not cache remote solutions.
    with grading_context.cache_level(
        grading_context.CacheLevel.NO_CACHE,
        when=lambda: is_path_remote(solution.code.path),
    ):
        return await steps_with_caching.run_coordinated(
            interactor_run_params,
            solution_run_params,
            sandbox=package.get_singleton_sandbox(),
            artifacts=grading_artifacts,
            dependency_cache=package.get_dependency_cache(),
            merged_capture=merged_capture_path,
            line_capture=line_capture,
        )



import pathlib

import pytest

from rbx.box.generation_schema import GenerationMetadata, GenerationTestcaseEntry
from rbx.box.schema import (
    CodeItem,
    ExpectedOutcome,
    Package,
    Solution,
    TaskType,
    Testcase,
)
from rbx.box.summary import (
    ContestProblemSummary,
    ProblemFlags,
    ProblemSummary,
    TestcaseCounts,
    count_testcases,
    get_contest_problem_summary,
    get_outcome_bucket,
    get_problem_flags,
    get_problem_summary,
    get_solution_counts,
)
from rbx.box.testcase_utils import TestcaseEntry


def _make_entry(group: str = 'tests', index: int = 0) -> GenerationTestcaseEntry:
    return GenerationTestcaseEntry(
        group_entry=TestcaseEntry(group=group, index=index),
        subgroup_entry=TestcaseEntry(group=group, index=index),
        metadata=GenerationMetadata(
            copied_to=Testcase(inputPath=pathlib.Path(f'{group}/{index}.in')),
        ),
    )


def _make_solution(
    outcome: ExpectedOutcome = ExpectedOutcome.ACCEPTED,
    path: str = 'sol.cpp',
) -> Solution:
    return Solution(path=pathlib.Path(path), outcome=outcome)


def _make_package(**kwargs) -> Package:
    defaults = {
        'name': 'test-problem',
        'timeLimit': 2000,
        'memoryLimit': 256,
    }
    defaults.update(kwargs)
    return Package(**defaults)


class TestCountTestcases:
    def test_empty(self):
        result = count_testcases([])
        assert result == TestcaseCounts(samples=0, hidden=0)

    def test_only_samples(self):
        entries = [_make_entry('samples', i) for i in range(3)]
        result = count_testcases(entries)
        assert result == TestcaseCounts(samples=3, hidden=0)

    def test_only_hidden(self):
        entries = [_make_entry('tests', i) for i in range(5)]
        result = count_testcases(entries)
        assert result == TestcaseCounts(samples=0, hidden=5)

    def test_mixed(self):
        entries = [
            _make_entry('samples', 0),
            _make_entry('samples', 1),
            _make_entry('tests', 0),
            _make_entry('corner', 0),
            _make_entry('random', 0),
        ]
        result = count_testcases(entries)
        assert result == TestcaseCounts(samples=2, hidden=3)


class TestGetOutcomeBucket:
    @pytest.mark.parametrize(
        'outcome,expected',
        [
            (ExpectedOutcome.ACCEPTED, ExpectedOutcome.ACCEPTED),
            (ExpectedOutcome.ACCEPTED_OR_TLE, ExpectedOutcome.ACCEPTED),
            (ExpectedOutcome.WRONG_ANSWER, ExpectedOutcome.WRONG_ANSWER),
            (ExpectedOutcome.INCORRECT, ExpectedOutcome.WRONG_ANSWER),
            (ExpectedOutcome.TIME_LIMIT_EXCEEDED, ExpectedOutcome.TIME_LIMIT_EXCEEDED),
            (ExpectedOutcome.TLE_OR_RTE, ExpectedOutcome.TIME_LIMIT_EXCEEDED),
            (ExpectedOutcome.RUNTIME_ERROR, ExpectedOutcome.RUNTIME_ERROR),
            (ExpectedOutcome.MEMORY_LIMIT_EXCEEDED, ExpectedOutcome.RUNTIME_ERROR),
            (ExpectedOutcome.OUTPUT_LIMIT_EXCEEDED, ExpectedOutcome.RUNTIME_ERROR),
            (ExpectedOutcome.JUDGE_FAILED, ExpectedOutcome.RUNTIME_ERROR),
            (ExpectedOutcome.COMPILATION_ERROR, ExpectedOutcome.RUNTIME_ERROR),
        ],
    )
    def test_known_outcomes(self, outcome, expected):
        assert get_outcome_bucket(outcome) == expected

    def test_any_returns_none(self):
        assert get_outcome_bucket(ExpectedOutcome.ANY) is None


class TestGetSolutionCountsExact:
    def test_empty(self):
        result = get_solution_counts([])
        assert all(v == 0 for v in result.values())
        # All ExpectedOutcome values should be keys.
        for outcome in ExpectedOutcome:
            assert outcome in result

    def test_counts_by_exact_outcome(self):
        solutions = [
            _make_solution(ExpectedOutcome.ACCEPTED, 'a.cpp'),
            _make_solution(ExpectedOutcome.ACCEPTED, 'b.cpp'),
            _make_solution(ExpectedOutcome.WRONG_ANSWER, 'wa.cpp'),
            _make_solution(ExpectedOutcome.ANY, 'any.cpp'),
        ]
        result = get_solution_counts(solutions)
        assert result[ExpectedOutcome.ACCEPTED] == 2
        assert result[ExpectedOutcome.WRONG_ANSWER] == 1
        assert result[ExpectedOutcome.ANY] == 1
        assert result[ExpectedOutcome.RUNTIME_ERROR] == 0


class TestGetSolutionCountsBucketed:
    def test_basic_bucketing(self):
        solutions = [
            _make_solution(ExpectedOutcome.ACCEPTED, 'a.cpp'),
            _make_solution(ExpectedOutcome.ACCEPTED_OR_TLE, 'b.cpp'),
            _make_solution(ExpectedOutcome.INCORRECT, 'c.cpp'),
            _make_solution(ExpectedOutcome.MEMORY_LIMIT_EXCEEDED, 'd.cpp'),
        ]
        result = get_solution_counts(solutions, bucketize=True)
        assert result[ExpectedOutcome.ACCEPTED] == 2
        assert result[ExpectedOutcome.WRONG_ANSWER] == 1
        assert result[ExpectedOutcome.RUNTIME_ERROR] == 1
        assert result[ExpectedOutcome.TIME_LIMIT_EXCEEDED] == 0

    def test_any_excluded_from_buckets(self):
        solutions = [
            _make_solution(ExpectedOutcome.ACCEPTED, 'a.cpp'),
            _make_solution(ExpectedOutcome.ANY, 'any.cpp'),
        ]
        result = get_solution_counts(solutions, bucketize=True)
        assert result[ExpectedOutcome.ACCEPTED] == 1
        # ANY should not be counted in any bucket.
        total_bucketed = sum(result.values())
        assert total_bucketed == 1

    def test_only_four_bucket_keys(self):
        result = get_solution_counts([], bucketize=True)
        assert set(result.keys()) == {
            ExpectedOutcome.ACCEPTED,
            ExpectedOutcome.WRONG_ANSWER,
            ExpectedOutcome.TIME_LIMIT_EXCEEDED,
            ExpectedOutcome.RUNTIME_ERROR,
        }


class TestGetProblemFlags:
    def test_defaults(self):
        pkg = _make_package()
        flags = get_problem_flags(pkg)
        assert flags == ProblemFlags(
            is_interactive=False, has_validator=False, has_custom_checker=False
        )

    def test_interactive(self):
        pkg = _make_package(type=TaskType.COMMUNICATION)
        flags = get_problem_flags(pkg)
        assert flags.is_interactive is True

    def test_validator(self):
        pkg = _make_package(validator=CodeItem(path=pathlib.Path('validator.cpp')))
        flags = get_problem_flags(pkg)
        assert flags.has_validator is True

    def test_checker(self):
        from rbx.box.schema import Checker

        pkg = _make_package(checker=Checker(path=pathlib.Path('checker.cpp')))
        flags = get_problem_flags(pkg)
        assert flags.has_custom_checker is True


class TestGetProblemSummary:
    def test_basic(self):
        pkg = _make_package(name='a-plus-b')
        solutions = [
            _make_solution(ExpectedOutcome.ACCEPTED, 'ac.cpp'),
            _make_solution(ExpectedOutcome.WRONG_ANSWER, 'wa.cpp'),
        ]
        entries = [_make_entry('samples', 0), _make_entry('tests', 0)]

        result = get_problem_summary(pkg, solutions, entries, short_name='A')

        assert isinstance(result, ProblemSummary)
        assert result.name == 'a-plus-b'
        assert result.short_name == 'A'
        assert result.time_limit_ms == 2000
        assert result.memory_limit_mb == 256
        assert result.testcase_counts == TestcaseCounts(samples=1, hidden=1)
        assert result.flags.is_interactive is False
        assert result.solution_counts[ExpectedOutcome.ACCEPTED] == 1
        assert result.solution_counts[ExpectedOutcome.WRONG_ANSWER] == 1
        assert len(result.solutions) == 2

    def test_no_short_name(self):
        pkg = _make_package()
        result = get_problem_summary(pkg, [], [])
        assert result.short_name is None


class TestGetContestProblemSummary:
    def test_basic(self):
        pkg = _make_package(name='sum')
        solutions = [
            _make_solution(ExpectedOutcome.ACCEPTED, 'ac.cpp'),
            _make_solution(ExpectedOutcome.ANY, 'any.cpp'),
            _make_solution(ExpectedOutcome.INCORRECT, 'wa.cpp'),
        ]
        entries = [_make_entry('samples', 0), _make_entry('tests', 0)]

        result = get_contest_problem_summary(pkg, solutions, entries, short_name='B')

        assert isinstance(result, ContestProblemSummary)
        assert result.short_name == 'B'
        assert result.name == 'sum'
        assert result.total_solutions == 3
        # ANY is excluded from buckets, so only 2 counted.
        total_bucketed = sum(result.solution_counts_bucketed.values())
        assert total_bucketed == 2
        assert result.solution_counts_bucketed[ExpectedOutcome.ACCEPTED] == 1
        assert result.solution_counts_bucketed[ExpectedOutcome.WRONG_ANSWER] == 1



"""LLM endpoints router for the wizard service."""

from __future__ import annotations

from typing import Optional

from agents import Runner
from fastapi import APIRouter, HTTPException
from fastapi.responses import PlainTextResponse
from pydantic import BaseModel, Field

from rbx.box.lang import code_to_lang
from rbx.box.wizard.agent_factories import (
    create_icpc_problem_reviewer,
    create_icpc_statement_language_reviewer,
)
from rbx.box.wizard.supported_models import DEFAULT_MODEL_NAME, SUPPORTED_MODELS

router = APIRouter(prefix='/llm', tags=['llm'])


class ModelsResponse(BaseModel):
    """Response for the models endpoint."""

    models: list[str] = Field(..., description='List of supported models')


@router.get('/models')
async def get_models() -> ModelsResponse:
    """Get the list of supported models."""

    return ModelsResponse(models=list(SUPPORTED_MODELS.keys()))


class ReviewRequest(BaseModel):
    """Input for the ICPC problem review endpoint."""

    statement: str = Field(..., description='Problem statement content')
    language: str = Field(..., description='Language of the statement, e.g., en, pt-br')
    validator: Optional[str] = Field(default=None, description='Validator source code')
    checker: Optional[str] = Field(default=None, description='Checker source code')
    interactor: Optional[str] = Field(
        default=None, description='Interactor source code'
    )
    model: Optional[str] = Field(
        default=None, description='Model to use for the review'
    )


## Removed structured ReviewResponse in favor of plain Markdown output


def _build_review_prompt(payload: ReviewRequest) -> str:
    """Create a concise, instruction-focused prompt for a Markdown review."""

    return (
        'You are an expert programming contest statement reviewer.\n\n'
        'Task: Review the problem statement (language: {language}), the checker (or interactor), and the validator.\n'
        'Identify whether constraints, limits, input/output formatting, and properties are consistent across the '
        'statement, checker, validator, and interactor.\n\n'
        'The statement is given in LaTeX. Pay attention to math expressions to identify constraints, and to the '
        'text itself to identify other issues.\n'
        'Example issues (non-exhaustive):\n'
        '- Constraints (bounds, ranges, etc.) not reflected equally in the statement and in testlib components.\n'
        '- Variables (vars.XXX) not used accordingly.\n\n'
        'Output format (Markdown):\n'
        '## ICPC Problem Review\n'
        '### Inconsistencies\n'
        '- List issues as bullet points. If none, write "None".\n'
        '### Validator issues\n'
        '- List issues as bullet points. If none, write "None".\n'
        '### Checker issues\n'
        '- List issues as bullet points. If none, write "None".\n'
        '### Interactor issues\n'
        '- List issues as bullet points. If none, write "None".\n'
        '- Include small code snippets when relevant.\n'
        '- When referring to code, explicitly mention file (checker/validator/interactor) and exact line numbers like `L12-L18`.\n\n'
        'Materials:\n'
        '[Statement]\n{statement}\n\n'
        '[Checker]\n{checker}\n\n'
        '[Interactor]\n{interactor}\n\n'
        '[Validator]\n{validator}\n'
    ).format(
        language=payload.language,
        statement=payload.statement,
        checker=payload.checker or 'N/A',
        validator=payload.validator or 'N/A',
        interactor=payload.interactor or 'N/A',
    )


@router.post('/review', response_class=PlainTextResponse)
async def review_endpoint(payload: ReviewRequest) -> PlainTextResponse:
    """Review an ICPC problem and return a Markdown report."""

    prompt = _build_review_prompt(payload)

    try:
        agent = create_icpc_problem_reviewer(payload.model or DEFAULT_MODEL_NAME)
        run_result = await Runner.run(agent, prompt)

        output = getattr(run_result, 'final_output', None)
        if isinstance(output, str):
            return PlainTextResponse(output, media_type='text/markdown')

        # Fallbacks
        if isinstance(run_result, str):
            return PlainTextResponse(run_result, media_type='text/markdown')

        text = str(output) if output is not None else str(run_result)
        return PlainTextResponse(text, media_type='text/markdown')
    except HTTPException:
        raise
    except Exception as exc:
        raise HTTPException(status_code=500, detail=str(exc)) from None


# ------------------------------
# Language-only statement review
# ------------------------------


class StatementLanguageReviewRequest(BaseModel):
    """Input for language-focused statement review.

    Only the statement text and its language code are required.
    """

    statement: str = Field(..., description='Problem statement content')
    language: str = Field(..., description='Language code, e.g., en, pt-br, es')
    model: Optional[str] = Field(
        default=None, description='Model to use for the review'
    )


def _build_language_review_prompt(payload: StatementLanguageReviewRequest) -> str:
    """Create a prompt that focuses purely on language quality.

    The assistant must analyze grammar, spelling/typos, orthography, clarity/style,
    terminology consistency, and major inconsistencies within the statement text.
    Output must be a Markdown string organized by categories with concise bullets
    and small snippet blocks for context.
    """

    return (
        'Review the following statement in the language "{language}". The lines of the statement are numbered from 1 to N '
        'starting from the next line, and the statement is given in LaTeX format.\n'
        '{statement}\n'
    ).format(language=code_to_lang(payload.language), statement=payload.statement)


@router.post('/review/statement', response_class=PlainTextResponse)
async def review_statement_language(
    payload: StatementLanguageReviewRequest,
) -> PlainTextResponse:
    """Review the statement's language and return a Markdown report."""

    prompt = _build_language_review_prompt(payload)

    try:
        agent = create_icpc_statement_language_reviewer(
            payload.model or DEFAULT_MODEL_NAME
        )
        run_result = await Runner.run(agent, prompt)

        output = getattr(run_result, 'final_output', None)
        if isinstance(output, str):
            return PlainTextResponse(output, media_type='text/markdown')

        # Fallbacks
        if isinstance(run_result, str):
            return PlainTextResponse(run_result, media_type='text/markdown')

        # If the model returns a structured object, convert to string
        text = str(output) if output is not None else str(run_result)
        return PlainTextResponse(text, media_type='text/markdown')
    except HTTPException:
        raise
    except Exception as exc:
        raise HTTPException(status_code=500, detail=str(exc)) from None
